"""Deprecated: use galileo.shared.filter instead of galileo.__future__.shared.filter."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.shared.filter is deprecated. Use galileo.shared.filter instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.shared.filter import (  # noqa: E402
    BooleanFilter,
    DateFilter,
    Filter,
    NumberFilter,
    TextFilter,
    boolean,
    date,
    number,
    text,
)

__all__ = ["BooleanFilter", "DateFilter", "Filter", "NumberFilter", "TextFilter", "boolean", "date", "number", "text"]



import datetime
from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.content_modality import ContentModality
from ..models.file_source import FileSource
from ..models.file_status import FileStatus
from ..types import UNSET, Unset

T = TypeVar("T", bound="FileMetadata")


@_attrs_define
class FileMetadata:
    """Enriched file metadata returned to UI/SDK.

    Contains presigned URLs and properties for displaying multimodal
    content in the Galileo console and SDKs.

    Attributes
    ----------
            file_id (str):
            modality (ContentModality): Classification of content modality
            source (FileSource): Source of the file data.
            status (FileStatus): Processing status of the file.
            content_type (Union[None, Unset, str]):
            url (Union[None, Unset, str]): Presigned S3 URL or external URL
            url_expires_at (Union[None, Unset, datetime.datetime]): Expiration time
            size_bytes (Union[None, Unset, int]):
            filename (Union[None, Unset, str]):
    """

    file_id: str
    modality: ContentModality
    source: FileSource
    status: FileStatus
    content_type: Union[None, Unset, str] = UNSET
    url: Union[None, Unset, str] = UNSET
    url_expires_at: Union[None, Unset, datetime.datetime] = UNSET
    size_bytes: Union[None, Unset, int] = UNSET
    filename: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        file_id = self.file_id

        modality = self.modality.value

        source = self.source.value

        status = self.status.value

        content_type: Union[None, Unset, str]
        content_type = UNSET if isinstance(self.content_type, Unset) else self.content_type

        url: Union[None, Unset, str]
        url = UNSET if isinstance(self.url, Unset) else self.url

        url_expires_at: Union[None, Unset, str]
        if isinstance(self.url_expires_at, Unset):
            url_expires_at = UNSET
        elif isinstance(self.url_expires_at, datetime.datetime):
            url_expires_at = self.url_expires_at.isoformat()
        else:
            url_expires_at = self.url_expires_at

        size_bytes: Union[None, Unset, int]
        size_bytes = UNSET if isinstance(self.size_bytes, Unset) else self.size_bytes

        filename: Union[None, Unset, str]
        filename = UNSET if isinstance(self.filename, Unset) else self.filename

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"file_id": file_id, "modality": modality, "source": source, "status": status})
        if content_type is not UNSET:
            field_dict["content_type"] = content_type
        if url is not UNSET:
            field_dict["url"] = url
        if url_expires_at is not UNSET:
            field_dict["url_expires_at"] = url_expires_at
        if size_bytes is not UNSET:
            field_dict["size_bytes"] = size_bytes
        if filename is not UNSET:
            field_dict["filename"] = filename

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        file_id = d.pop("file_id")

        modality = ContentModality(d.pop("modality"))

        source = FileSource(d.pop("source"))

        status = FileStatus(d.pop("status"))

        def _parse_content_type(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        content_type = _parse_content_type(d.pop("content_type", UNSET))

        def _parse_url(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        url = _parse_url(d.pop("url", UNSET))

        def _parse_url_expires_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                return isoparse(data)

            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        url_expires_at = _parse_url_expires_at(d.pop("url_expires_at", UNSET))

        def _parse_size_bytes(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        size_bytes = _parse_size_bytes(d.pop("size_bytes", UNSET))

        def _parse_filename(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        filename = _parse_filename(d.pop("filename", UNSET))

        file_metadata = cls(
            file_id=file_id,
            modality=modality,
            source=source,
            status=status,
            content_type=content_type,
            url=url,
            url_expires_at=url_expires_at,
            size_bytes=size_bytes,
            filename=filename,
        )

        file_metadata.additional_properties = d
        return file_metadata

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class FileStatus(str, Enum):
    COMPLETE = "complete"
    FAILED = "failed"
    NOT_UPLOADED = "not_uploaded"
    PENDING = "pending"

    def __str__(self) -> str:
        return str(self.value)



"""Deprecated: use galileo.shared.base instead of galileo.__future__.shared.base."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.shared.base is deprecated. Use galileo.shared.base instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.shared.base import StateManagementMixin, SyncState  # noqa: E402

__all__ = ["StateManagementMixin", "SyncState"]



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.annotation_like_dislike_aggregate import AnnotationLikeDislikeAggregate
    from ..models.annotation_score_aggregate import AnnotationScoreAggregate
    from ..models.annotation_star_aggregate import AnnotationStarAggregate
    from ..models.annotation_tags_aggregate import AnnotationTagsAggregate
    from ..models.annotation_text_aggregate import AnnotationTextAggregate


T = TypeVar("T", bound="AnnotationAggregate")


@_attrs_define
class AnnotationAggregate:
    """
    Attributes
    ----------
        aggregate (Union['AnnotationLikeDislikeAggregate', 'AnnotationScoreAggregate', 'AnnotationStarAggregate',
            'AnnotationTagsAggregate', 'AnnotationTextAggregate']):
    """

    aggregate: Union[
        "AnnotationLikeDislikeAggregate",
        "AnnotationScoreAggregate",
        "AnnotationStarAggregate",
        "AnnotationTagsAggregate",
        "AnnotationTextAggregate",
    ]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.annotation_like_dislike_aggregate import AnnotationLikeDislikeAggregate
        from ..models.annotation_score_aggregate import AnnotationScoreAggregate
        from ..models.annotation_star_aggregate import AnnotationStarAggregate
        from ..models.annotation_tags_aggregate import AnnotationTagsAggregate

        aggregate: dict[str, Any]
        if isinstance(
            self.aggregate,
            (
                AnnotationLikeDislikeAggregate,
                AnnotationStarAggregate,
                AnnotationScoreAggregate,
                AnnotationTagsAggregate,
            ),
        ):
            aggregate = self.aggregate.to_dict()
        else:
            aggregate = self.aggregate.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"aggregate": aggregate})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.annotation_like_dislike_aggregate import AnnotationLikeDislikeAggregate
        from ..models.annotation_score_aggregate import AnnotationScoreAggregate
        from ..models.annotation_star_aggregate import AnnotationStarAggregate
        from ..models.annotation_tags_aggregate import AnnotationTagsAggregate
        from ..models.annotation_text_aggregate import AnnotationTextAggregate

        d = dict(src_dict)

        def _parse_aggregate(
            data: object,
        ) -> Union[
            "AnnotationLikeDislikeAggregate",
            "AnnotationScoreAggregate",
            "AnnotationStarAggregate",
            "AnnotationTagsAggregate",
            "AnnotationTextAggregate",
        ]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return AnnotationLikeDislikeAggregate.from_dict(data)

            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return AnnotationStarAggregate.from_dict(data)

            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return AnnotationScoreAggregate.from_dict(data)

            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return AnnotationTagsAggregate.from_dict(data)

            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            return AnnotationTextAggregate.from_dict(data)

        aggregate = _parse_aggregate(d.pop("aggregate"))

        annotation_aggregate = cls(aggregate=aggregate)

        annotation_aggregate.additional_properties = d
        return annotation_aggregate

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Deprecated: use galileo.shared.exceptions instead of galileo.__future__.shared.exceptions."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.shared.exceptions is deprecated. Use galileo.shared.exceptions instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.shared.exceptions import (  # noqa: E402
    APIError,
    ConfigurationError,
    GalileoFutureError,
    IntegrationNotConfiguredError,
    ResourceConflictError,
    ResourceNotFoundError,
    SyncError,
    ValidationError,
)

__all__ = [
    "APIError",
    "ConfigurationError",
    "GalileoFutureError",
    "IntegrationNotConfiguredError",
    "ResourceConflictError",
    "ResourceNotFoundError",
    "SyncError",
    "ValidationError",
]



"""Deprecated: use galileo.prompt instead of galileo.__future__.prompt."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.prompt is deprecated. Use galileo.prompt instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.prompt import Prompt, PromptVersion, _parse_template_to_messages  # noqa: E402

__all__ = ["Prompt", "PromptVersion", "_parse_template_to_messages"]



"""Deprecated: use galileo.integration instead of galileo.__future__.integration."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.integration is deprecated. Use galileo.integration instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.integration import Integration  # noqa: E402

__all__ = ["Integration"]



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="VertexAIGCSConfig")


@_attrs_define
class VertexAIGCSConfig:
    """Configuration for GCS file uploads in Vertex AI.

    Attributes
    ----------
        service_account_credentials (str):
        bucket_name (str):
        object_path_prefix (str):
    """

    service_account_credentials: str
    bucket_name: str
    object_path_prefix: str
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        service_account_credentials = self.service_account_credentials

        bucket_name = self.bucket_name

        object_path_prefix = self.object_path_prefix

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "service_account_credentials": service_account_credentials,
                "bucket_name": bucket_name,
                "object_path_prefix": object_path_prefix,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        service_account_credentials = d.pop("service_account_credentials")

        bucket_name = d.pop("bucket_name")

        object_path_prefix = d.pop("object_path_prefix")

        vertex_aigcs_config = cls(
            service_account_credentials=service_account_credentials,
            bucket_name=bucket_name,
            object_path_prefix=object_path_prefix,
        )

        vertex_aigcs_config.additional_properties = d
        return vertex_aigcs_config

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Deprecated: use galileo.experiment instead of galileo.__future__.experiment."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.experiment is deprecated. Use galileo.experiment instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.experiment import Experiment  # noqa: E402

__all__ = ["Experiment"]



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.error_severity import ErrorSeverity
from ..models.error_type import ErrorType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.standard_error_context import StandardErrorContext


T = TypeVar("T", bound="StandardError")


@_attrs_define
class StandardError:
    """
    Attributes
    ----------
        error_code (int):
        error_type (ErrorType):
        error_group (str):
        severity (ErrorSeverity): Error severity levels for catalog entries.
        message (str):
        user_action (Union[None, Unset, str]):
        documentation_link (Union[None, Unset, str]):
        retriable (Union[Unset, bool]):  Default: False.
        blocking (Union[Unset, bool]):  Default: False.
        http_status_code (Union[None, Unset, int]):
        source_service (Union[None, Unset, str]):
        context (Union[Unset, StandardErrorContext]):
    """

    error_code: int
    error_type: ErrorType
    error_group: str
    severity: ErrorSeverity
    message: str
    user_action: Union[None, Unset, str] = UNSET
    documentation_link: Union[None, Unset, str] = UNSET
    retriable: Union[Unset, bool] = False
    blocking: Union[Unset, bool] = False
    http_status_code: Union[None, Unset, int] = UNSET
    source_service: Union[None, Unset, str] = UNSET
    context: Union[Unset, "StandardErrorContext"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        error_code = self.error_code

        error_type = self.error_type.value

        error_group = self.error_group

        severity = self.severity.value

        message = self.message

        user_action: Union[None, Unset, str]
        user_action = UNSET if isinstance(self.user_action, Unset) else self.user_action

        documentation_link: Union[None, Unset, str]
        documentation_link = UNSET if isinstance(self.documentation_link, Unset) else self.documentation_link

        retriable = self.retriable

        blocking = self.blocking

        http_status_code: Union[None, Unset, int]
        http_status_code = UNSET if isinstance(self.http_status_code, Unset) else self.http_status_code

        source_service: Union[None, Unset, str]
        source_service = UNSET if isinstance(self.source_service, Unset) else self.source_service

        context: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.context, Unset):
            context = self.context.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "error_code": error_code,
                "error_type": error_type,
                "error_group": error_group,
                "severity": severity,
                "message": message,
            }
        )
        if user_action is not UNSET:
            field_dict["user_action"] = user_action
        if documentation_link is not UNSET:
            field_dict["documentation_link"] = documentation_link
        if retriable is not UNSET:
            field_dict["retriable"] = retriable
        if blocking is not UNSET:
            field_dict["blocking"] = blocking
        if http_status_code is not UNSET:
            field_dict["http_status_code"] = http_status_code
        if source_service is not UNSET:
            field_dict["source_service"] = source_service
        if context is not UNSET:
            field_dict["context"] = context

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.standard_error_context import StandardErrorContext

        d = dict(src_dict)
        error_code = d.pop("error_code")

        error_type = ErrorType(d.pop("error_type"))

        error_group = d.pop("error_group")

        severity = ErrorSeverity(d.pop("severity"))

        message = d.pop("message")

        def _parse_user_action(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        user_action = _parse_user_action(d.pop("user_action", UNSET))

        def _parse_documentation_link(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        documentation_link = _parse_documentation_link(d.pop("documentation_link", UNSET))

        retriable = d.pop("retriable", UNSET)

        blocking = d.pop("blocking", UNSET)

        def _parse_http_status_code(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        http_status_code = _parse_http_status_code(d.pop("http_status_code", UNSET))

        def _parse_source_service(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        source_service = _parse_source_service(d.pop("source_service", UNSET))

        _context = d.pop("context", UNSET)
        context: Union[Unset, StandardErrorContext]
        context = UNSET if isinstance(_context, Unset) else StandardErrorContext.from_dict(_context)

        standard_error = cls(
            error_code=error_code,
            error_type=error_type,
            error_group=error_group,
            severity=severity,
            message=message,
            user_action=user_action,
            documentation_link=documentation_link,
            retriable=retriable,
            blocking=blocking,
            http_status_code=http_status_code,
            source_service=source_service,
            context=context,
        )

        standard_error.additional_properties = d
        return standard_error

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Deprecated: use galileo.shared.experiment_result instead of galileo.__future__.shared.experiment_result."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.shared.experiment_result is deprecated. "
    "Use galileo.shared.experiment_result instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.shared.experiment_result import (  # noqa: E402
    ExperimentPhaseInfo,
    ExperimentRunResult,
    ExperimentStatusInfo,
)

__all__ = ["ExperimentPhaseInfo", "ExperimentRunResult", "ExperimentStatusInfo"]



from collections.abc import Mapping
from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.content_modality import ContentModality
from ..models.multimodal_capability import MultimodalCapability
from ..types import UNSET, Unset

T = TypeVar("T", bound="ModelProperties")


@_attrs_define
class ModelProperties:
    """
    Attributes
    ----------
        alias (str):
        name (str):
        input_modalities (list[ContentModality]):
        multimodal_capabilities (Union[Unset, list[MultimodalCapability]]):
    """

    alias: str
    name: str
    input_modalities: list[ContentModality]
    multimodal_capabilities: Union[Unset, list[MultimodalCapability]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        alias = self.alias

        name = self.name

        input_modalities = []
        for input_modalities_item_data in self.input_modalities:
            input_modalities_item = input_modalities_item_data.value
            input_modalities.append(input_modalities_item)

        multimodal_capabilities: Union[Unset, list[str]] = UNSET
        if not isinstance(self.multimodal_capabilities, Unset):
            multimodal_capabilities = []
            for multimodal_capabilities_item_data in self.multimodal_capabilities:
                multimodal_capabilities_item = multimodal_capabilities_item_data.value
                multimodal_capabilities.append(multimodal_capabilities_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"alias": alias, "name": name, "input_modalities": input_modalities})
        if multimodal_capabilities is not UNSET:
            field_dict["multimodal_capabilities"] = multimodal_capabilities

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        alias = d.pop("alias")

        name = d.pop("name")

        input_modalities = []
        _input_modalities = d.pop("input_modalities")
        for input_modalities_item_data in _input_modalities:
            input_modalities_item = ContentModality(input_modalities_item_data)

            input_modalities.append(input_modalities_item)

        multimodal_capabilities = []
        _multimodal_capabilities = d.pop("multimodal_capabilities", UNSET)
        for multimodal_capabilities_item_data in _multimodal_capabilities or []:
            multimodal_capabilities_item = MultimodalCapability(multimodal_capabilities_item_data)

            multimodal_capabilities.append(multimodal_capabilities_item)

        model_properties = cls(
            alias=alias, name=name, input_modalities=input_modalities, multimodal_capabilities=multimodal_capabilities
        )

        model_properties.additional_properties = d
        return model_properties

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class FileSource(str, Enum):
    ASSEMBLED_STREAM = "assembled_stream"
    DIRECT_UPLOAD = "direct_upload"
    EXTERNAL_FILES_API = "external_files_api"
    EXTERNAL_URL = "external_url"

    def __str__(self) -> str:
        return str(self.value)



"""Base classes for lifecycle and state management."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any

logger = logging.getLogger(__name__)

_SENTINEL = object()


class SyncState(Enum):
    """
    Enumeration of possible synchronization states for business objects.

    Attributes
    ----------
        LOCAL_ONLY: Object exists only in memory, not yet persisted remotely.
        SYNCED: Local and remote states match.
        DIRTY: Local changes exist that have not been saved.
        FAILED_SYNC: Last attempt to sync failed.
        DELETED: Object was deleted remotely, local object still exists.
    """

    LOCAL_ONLY = "local_only"
    SYNCED = "synced"
    DIRTY = "dirty"
    FAILED_SYNC = "failed_sync"
    DELETED = "deleted"


class StateManagementMixin(ABC):
    """
    Base mixin for business objects providing lifecycle state management.

    This mixin provides state tracking and helper methods for objects that
    need to synchronize between local and remote (API) states.

    Subclasses may declare a ``_TRACKED_FIELDS`` frozenset to enable automatic
    dirty-tracking: any assignment to a tracked field on a SYNCED object will
    transition it to DIRTY so that callers know a ``save()`` is needed.

    Attributes
    ----------
        _sync_state: Current synchronization state of the object.
        _last_error: Last error encountered during synchronization (optional).
        _TRACKED_FIELDS: frozenset of attribute names whose mutations trigger
            SYNCED → DIRTY transitions.
    """

    _TRACKED_FIELDS: frozenset[str] = frozenset()

    def __init__(self) -> None:
        """Initialize the mixin with default state."""
        self._sync_state: SyncState = SyncState.LOCAL_ONLY
        self._last_error: Exception | None = None

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Override attribute setting to auto-transition SYNCED → DIRTY.

        Only attributes listed in ``type(self)._TRACKED_FIELDS`` trigger the
        transition, and only when the incoming value actually differs from the
        current value (same-value assignments are no-ops).

        Parameters
        ----------
            name: Attribute name being set.
            value: New value for the attribute.
        """
        tracked = type(self)._TRACKED_FIELDS
        if name in tracked:
            current = object.__getattribute__(self, name) if hasattr(self, name) else _SENTINEL
            if current is not _SENTINEL and current != value:
                # Only transition from SYNCED; LOCAL_ONLY/DIRTY/etc. are unaffected
                try:
                    sync_state = object.__getattribute__(self, "_sync_state")
                except AttributeError:
                    sync_state = None
                if sync_state == SyncState.SYNCED:
                    object.__setattr__(self, "_sync_state", SyncState.DIRTY)
        object.__setattr__(self, name, value)

    def _sync_attrs(self, **attrs: Any) -> None:
        """
        Bulk-set attributes without triggering dirty-tracking.

        Use this inside ``create()``, ``refresh()``, and ``save()`` to write
        API-sourced values back onto the instance without transitioning state.

        Parameters
        ----------
            **attrs: Keyword arguments mapping attribute names to their new values.
        """
        for key, value in attrs.items():
            object.__setattr__(self, key, value)

    @property
    def sync_state(self) -> SyncState:
        """
        Get the current synchronization state.

        Returns
        -------
            SyncState: The current state of the object.
        """
        return self._sync_state

    def _set_state(self, state: SyncState, error: Exception | None = None) -> None:
        """
        Set the synchronization state and optionally store an error.

        Args:
            state: The new synchronization state.
            error: Optional exception that caused the state transition.
        """
        old_state = self._sync_state
        self._sync_state = state
        self._last_error = error

        # Log state transitions at DEBUG level
        obj_name = self.__class__.__name__
        obj_id = getattr(self, "id", "unknown")
        obj_name_attr = getattr(self, "name", "unknown")

        if error:
            logger.debug(
                f"{obj_name} state transition: {old_state.value} → {state.value} "
                f"(id={obj_id}, name={obj_name_attr}, error={str(error)[:100]})"
            )
        else:
            logger.debug(
                f"{obj_name} state transition: {old_state.value} → {state.value} (id={obj_id}, name={obj_name_attr})"
            )

    def is_synced(self) -> bool:
        """
        Check if the object is synchronized with the remote state.

        Returns
        -------
            bool: True if the object is synced, False otherwise.
        """
        return self._sync_state == SyncState.SYNCED

    def is_dirty(self) -> bool:
        """
        Check if the object has unsaved local changes.

        Returns
        -------
            bool: True if the object has unsaved changes, False otherwise.
        """
        return self._sync_state == SyncState.DIRTY

    def is_local_only(self) -> bool:
        """
        Check if the object exists only locally.

        Returns
        -------
            bool: True if the object has not been persisted, False otherwise.
        """
        return self._sync_state == SyncState.LOCAL_ONLY

    def has_failed(self) -> bool:
        """
        Check if the last synchronization attempt failed.

        Returns
        -------
            bool: True if the last sync failed, False otherwise.
        """
        return self._sync_state == SyncState.FAILED_SYNC

    def is_deleted(self) -> bool:
        """
        Check if the object has been deleted remotely.

        Returns
        -------
            bool: True if the object was deleted, False otherwise.
        """
        return self._sync_state == SyncState.DELETED

    @abstractmethod
    def refresh(self) -> None:
        """
        Refresh the object state from the remote API.

        This method should:
        1. Fetch the latest state from the API
        2. Update all local attributes
        3. Set the state to SYNCED on success

        Raises
        ------
            NotImplementedError: If not implemented by subclass.
        """
        raise NotImplementedError("Subclasses must implement refresh()")



# mypy: disable-error-code=syntax
# We need to ignore syntax errors until https://github.com/python/mypy/issues/17535 is resolved.
import os
from typing import Any, ClassVar, Optional

from pydantic_core import Url

from galileo.constants import DEFAULT_CONSOLE_URL
from galileo.shared.exceptions import ConfigurationError
from galileo_core.schemas.base_config import GalileoConfig


class GalileoPythonConfig(GalileoConfig):
    # Config file for this project.
    config_filename: str = "galileo-python-config.json"
    console_url: Url = DEFAULT_CONSOLE_URL

    _instance: ClassVar[Optional["GalileoPythonConfig"]] = None

    def reset(self) -> None:
        GalileoPythonConfig._instance = None
        super().reset()

    @classmethod
    def get(cls, **kwargs: Any) -> "GalileoPythonConfig":
        # Check for any supported auth method before raising.
        # TODO: also check GALILEO_USERNAME / GALILEO_PASSWORD / SSO env vars once alternative auth is surfaced.
        _AUTH_KWARGS = ("api_key", "username", "password", "sso_id_token", "sso_provider")
        if (
            cls._instance is None
            and not any(kwargs.get(k) for k in _AUTH_KWARGS)
            and not os.environ.get("GALILEO_API_KEY")
        ):
            raise ConfigurationError(
                "GALILEO_API_KEY not detected. "
                "Please set the GALILEO_API_KEY environment variable or pass api_key= as a keyword argument. "
                "See https://docs.galileo.ai for setup instructions."
            )
        cls._instance = cls._get(cls._instance, **kwargs)
        assert cls._instance is not None, "Failed to initialize GalileoPythonConfig"
        return cls._instance



import hashlib
import json
import logging
from datetime import datetime
from typing import Any, Optional, Union
from uuid import UUID

from packaging.version import Version

from galileo.handlers.base_handler import GalileoBaseHandler
from galileo.logger import GalileoLogger
from galileo.schema.handlers import NodeType
from galileo.utils.serialization import serialize_to_str

_logger = logging.getLogger(__name__)

# Lazy import state — crewai wraps sys.stdout/sys.stderr at import time, which
# causes pytest-xdist worker hangs.  All crewai/litellm imports are deferred to
# _resolve_crewai_imports(), called once from CrewAIEventListener.__init__.
CREWAI_AVAILABLE = False
CREWAI_EVENTS_MODULE_AVAILABLE = False
LITE_LLM_AVAILABLE = False
litellm = None
_crewai_imports_resolved = False


def _resolve_crewai_imports() -> None:
    """Import crewai and litellm on first use, populating module-level globals."""
    global CREWAI_AVAILABLE, CREWAI_EVENTS_MODULE_AVAILABLE  # noqa: PLW0603
    global LITE_LLM_AVAILABLE, litellm, _crewai_imports_resolved  # noqa: PLW0603

    if _crewai_imports_resolved:
        return
    _crewai_imports_resolved = True

    try:
        from crewai import __version__ as _crewai_version_str

        crewai_version = Version(_crewai_version_str)
        CREWAI_EVENTS_MODULE_AVAILABLE = Version("0.177.0") <= crewai_version

        CREWAI_AVAILABLE = True
    except ImportError:
        _logger.warning("CrewAI not available, using stubs")

    try:
        import litellm as _litellm

        litellm = _litellm
        LITE_LLM_AVAILABLE = True
    except ImportError:
        _logger.warning("LiteLLM not available, using stubs")
        litellm = None


class CrewAIEventListener:
    """
    CrewAI event listener for logging traces to the Galileo platform.

    Attributes
    ----------
    _handler : GalileoBaseHandler
        The handler for managing the trace.
    """

    def __init__(
        self,
        galileo_logger: Optional[GalileoLogger] = None,
        start_new_trace: bool = True,
        flush_on_crew_completed: bool = True,
    ):
        _resolve_crewai_imports()

        self._handler = GalileoBaseHandler(
            flush_on_chain_end=flush_on_crew_completed,
            start_new_trace=start_new_trace,
            galileo_logger=galileo_logger,
            integration="crewai",
        )

        if CREWAI_AVAILABLE:
            # Register with the event bus directly instead of inheriting from
            # BaseEventListener — dynamic __bases__ assignment is rejected by
            # CPython when the deallocator differs from object's.
            try:
                if CREWAI_EVENTS_MODULE_AVAILABLE:
                    from crewai.events import crewai_event_bus  # pyright: ignore[reportMissingImports]
                else:
                    from crewai.utilities.events import crewai_event_bus  # pyright: ignore[reportMissingImports]
                self.setup_listeners(crewai_event_bus)
            except ImportError:
                _logger.warning("Could not import crewai event bus, skipping listener setup")

        if LITE_LLM_AVAILABLE and litellm is not None:
            if not litellm.success_callback:
                litellm.success_callback = []
            litellm.success_callback.append(self.lite_llm_usage_callback)

    def setup_listeners(self, crewai_event_bus: Any) -> None:
        """Setup event listeners for CrewAI events."""
        if not CREWAI_AVAILABLE:
            _logger.warning("CrewAI not available, skipping event listener setup")
            return

        # Lazy-import event types — crewai is guaranteed to be in sys.modules at
        # this point because _resolve_crewai_imports() ran in __init__.
        if CREWAI_EVENTS_MODULE_AVAILABLE:
            from crewai.events.types.agent_events import (  # pyright: ignore[reportMissingImports]
                AgentExecutionCompletedEvent,
                AgentExecutionErrorEvent,
                AgentExecutionStartedEvent,
            )
            from crewai.events.types.crew_events import (  # pyright: ignore[reportMissingImports]
                CrewKickoffCompletedEvent,
                CrewKickoffFailedEvent,
                CrewKickoffStartedEvent,
            )
            from crewai.events.types.llm_events import (  # pyright: ignore[reportMissingImports]
                LLMCallCompletedEvent,
                LLMCallFailedEvent,
                LLMCallStartedEvent,
            )
            from crewai.events.types.memory_events import (  # pyright: ignore[reportMissingImports]
                MemoryQueryCompletedEvent,
                MemoryQueryFailedEvent,
                MemoryQueryStartedEvent,
                MemoryRetrievalCompletedEvent,
                MemoryRetrievalStartedEvent,
                MemorySaveCompletedEvent,
                MemorySaveFailedEvent,
                MemorySaveStartedEvent,
            )

            try:
                from crewai.events.types.memory_events import (  # pyright: ignore[reportMissingImports]
                    MemoryRetrievalFailedEvent,
                )
            except ImportError:
                MemoryRetrievalFailedEvent = None

            from crewai.events.types.task_events import (  # pyright: ignore[reportMissingImports]
                TaskCompletedEvent,
                TaskFailedEvent,
                TaskStartedEvent,
            )
            from crewai.events.types.tool_usage_events import (  # pyright: ignore[reportMissingImports]
                ToolUsageErrorEvent,
                ToolUsageFinishedEvent,
                ToolUsageStartedEvent,
            )
        else:
            from crewai.utilities.events.agent_events import (  # pyright: ignore[reportMissingImports]
                AgentExecutionCompletedEvent,
                AgentExecutionErrorEvent,
                AgentExecutionStartedEvent,
            )
            from crewai.utilities.events.crew_events import (  # pyright: ignore[reportMissingImports]
                CrewKickoffCompletedEvent,
                CrewKickoffFailedEvent,
                CrewKickoffStartedEvent,
            )
            from crewai.utilities.events.llm_events import (  # pyright: ignore[reportMissingImports]
                LLMCallCompletedEvent,
                LLMCallFailedEvent,
                LLMCallStartedEvent,
            )
            from crewai.utilities.events.task_events import (  # pyright: ignore[reportMissingImports]
                TaskCompletedEvent,
                TaskFailedEvent,
                TaskStartedEvent,
            )
            from crewai.utilities.events.tool_usage_events import (  # pyright: ignore[reportMissingImports]
                ToolUsageErrorEvent,
                ToolUsageFinishedEvent,
                ToolUsageStartedEvent,
            )

            MemoryRetrievalFailedEvent = None

        # Crew event handlers
        @crewai_event_bus.on(CrewKickoffStartedEvent)
        def on_crew_kickoff_started(source: Any, event: CrewKickoffStartedEvent) -> None:
            self._handle_crew_kickoff_started(source, event)

        @crewai_event_bus.on(CrewKickoffCompletedEvent)
        def on_crew_kickoff_completed(source: Any, event: CrewKickoffCompletedEvent) -> None:
            self._handle_crew_kickoff_completed(source, event)

        @crewai_event_bus.on(CrewKickoffFailedEvent)
        def on_crew_kickoff_failed(source: Any, event: CrewKickoffFailedEvent) -> None:
            self._handle_crew_kickoff_failed(source, event)

        # Agent event handlers
        @crewai_event_bus.on(AgentExecutionStartedEvent)
        def on_agent_execution_started(source: Any, event: AgentExecutionStartedEvent) -> None:
            self._handle_agent_execution_started(source, event)

        @crewai_event_bus.on(AgentExecutionCompletedEvent)
        def on_agent_execution_completed(source: Any, event: AgentExecutionCompletedEvent) -> None:
            self._handle_agent_execution_completed(source, event)

        @crewai_event_bus.on(AgentExecutionErrorEvent)
        def on_agent_execution_error(source: Any, event: AgentExecutionErrorEvent) -> None:
            self._handle_agent_execution_error(source, event)

        # Task event handlers
        @crewai_event_bus.on(TaskStartedEvent)
        def on_task_started(source: Any, event: TaskStartedEvent) -> None:
            self._handle_task_started(source, event)

        @crewai_event_bus.on(TaskCompletedEvent)
        def on_task_completed(source: Any, event: TaskCompletedEvent) -> None:
            self._handle_task_completed(source, event)

        @crewai_event_bus.on(TaskFailedEvent)
        def on_task_failed(source: Any, event: TaskFailedEvent) -> None:
            self._handle_task_failed(source, event)

        # Tool event handlers
        @crewai_event_bus.on(ToolUsageStartedEvent)
        def on_tool_usage_started(source: Any, event: ToolUsageStartedEvent) -> None:
            self._handle_tool_usage_started(source, event)

        @crewai_event_bus.on(ToolUsageFinishedEvent)
        def on_tool_usage_finished(source: Any, event: ToolUsageFinishedEvent) -> None:
            self._handle_tool_usage_finished(source, event)

        @crewai_event_bus.on(ToolUsageErrorEvent)
        def on_tool_usage_error(source: Any, event: ToolUsageErrorEvent) -> None:
            self._handle_tool_usage_error(source, event)

        # LLM event handlers
        @crewai_event_bus.on(LLMCallStartedEvent)
        def on_llm_call_started(source: Any, event: LLMCallStartedEvent) -> None:
            self._handle_llm_call_started(source, event)

        @crewai_event_bus.on(LLMCallCompletedEvent)
        def on_llm_call_completed(source: Any, event: LLMCallCompletedEvent) -> None:
            self._handle_llm_call_completed(source, event)

        @crewai_event_bus.on(LLMCallFailedEvent)
        def on_llm_call_failed(source: Any, event: LLMCallFailedEvent) -> None:
            self._handle_llm_call_failed(source, event)

        # Memory event handlers (only available in CrewAI >= 0.177.0)
        if CREWAI_EVENTS_MODULE_AVAILABLE:

            @crewai_event_bus.on(MemoryQueryStartedEvent)
            def on_memory_query_started(source: Any, event: MemoryQueryStartedEvent) -> None:
                self._handle_memory_query_started(source, event)

            @crewai_event_bus.on(MemoryQueryCompletedEvent)
            def on_memory_query_completed(source: Any, event: MemoryQueryCompletedEvent) -> None:
                self._handle_memory_query_completed(source, event)

            @crewai_event_bus.on(MemoryQueryFailedEvent)
            def on_memory_query_failed(source: Any, event: MemoryQueryFailedEvent) -> None:
                self._handle_memory_query_failed(source, event)

            @crewai_event_bus.on(MemorySaveStartedEvent)
            def on_memory_save_started(source: Any, event: MemorySaveStartedEvent) -> None:
                self._handle_memory_save_started(source, event)

            @crewai_event_bus.on(MemorySaveCompletedEvent)
            def on_memory_save_completed(source: Any, event: MemorySaveCompletedEvent) -> None:
                self._handle_memory_save_completed(source, event)

            @crewai_event_bus.on(MemorySaveFailedEvent)
            def on_memory_save_failed(source: Any, event: MemorySaveFailedEvent) -> None:
                self._handle_memory_save_failed(source, event)

            @crewai_event_bus.on(MemoryRetrievalStartedEvent)
            def on_memory_retrieval_started(source: Any, event: MemoryRetrievalStartedEvent) -> None:
                self._handle_memory_retrieval_started(source, event)

            @crewai_event_bus.on(MemoryRetrievalCompletedEvent)
            def on_memory_retrieval_completed(source: Any, event: MemoryRetrievalCompletedEvent) -> None:
                self._handle_memory_retrieval_completed(source, event)

            if MemoryRetrievalFailedEvent is not None:

                @crewai_event_bus.on(MemoryRetrievalFailedEvent)
                def on_memory_retrieval_failed(source: Any, event: MemoryRetrievalFailedEvent) -> None:
                    self._handle_memory_retrieval_failed(source, event)

    @staticmethod
    def _hash_to_uuid(value: str) -> UUID:
        """Hash a string to a deterministic UUID."""
        return UUID(hashlib.md5(value.encode()).hexdigest())

    def _generate_run_id(self, source: Any, event: Any) -> UUID:
        """Generate a consistent UUID for event tracing."""
        # 1. Memory event specific ID generation
        if hasattr(event, "query") and getattr(event, "type", "").startswith("memory_query"):  # Memory query events
            source_type = event.source_type if hasattr(event, "source_type") else ""
            return self._hash_to_uuid(
                f"memory_query_{event.query}_{source_type}_{getattr(event, 'agent_id', '')}_{getattr(event, 'limit', '')}_{getattr(event, 'score_threshold', '')}"
            )

        if hasattr(event, "value") and getattr(event, "type", "").startswith("memory_save"):  # Memory save events
            return self._hash_to_uuid(
                f"memory_save_{getattr(event, 'agent_id', '')}_{getattr(event, 'task_id', '')}_{getattr(event, 'agent_role', '')}"
            )

        if hasattr(event, "memory_content") or getattr(event, "type", "").startswith(
            "memory_retrieval"
        ):  # Memory retrieval events
            # Use consistent fields for all retrieval events (started/completed/failed) - exclude memory_content to ensure same ID
            return self._hash_to_uuid(
                f"memory_retrieval_{getattr(event, 'task_id', '')}_{getattr(event, 'agent_id', '')}"
            )

        # 2. event.call_id — LLM events in crewAI >= 1.0
        call_id = getattr(event, "call_id", None)
        if call_id:
            return self._hash_to_uuid(f"call_{call_id}")

        # 3. source.id — still works for Crew/Task sources
        if hasattr(source, "id") and source.id:
            return source.id

        # 4. event.messages hash — LLM fallback for crewAI < 1.0
        if hasattr(event, "messages") and event.messages is not None:
            messages = json.dumps(event.to_json().get("messages", []))
            return self._hash_to_uuid(messages)

        # 5. Tool events — use crewAI's event scope tracking for correlation (crewAI >= 1.0)
        if hasattr(event, "tool_args"):
            # crewAI < 1.0: source is a Tool instance with .id, already handled by step 3.
            # Fallback to tool_args hash for other cases.
            tool_name = getattr(event, "tool_name", "")
            if isinstance(event.tool_args, str):
                tool_args = event.tool_args
            else:
                tool_args = json.dumps(event.tool_args) if isinstance(event.tool_args, dict) else str(event.tool_args)
            agent_id = getattr(event, "agent_id", "") or ""
            return self._hash_to_uuid(f"{tool_name}_{tool_args}_{agent_id}")

        # 6. event.agent.id — Agent events in crewAI >= 1.0 (source no longer has .id)
        event_agent = getattr(event, "agent", None)
        if event_agent is not None and getattr(event_agent, "id", None):
            return event_agent.id

        # 7. dict messages — lite_llm callback
        if isinstance(event, dict) and "messages" in event:
            messages = json.dumps(event["messages"])
            return self._hash_to_uuid(messages)

        # 8. Generic fallback
        return self._hash_to_uuid(
            f"{getattr(event, 'crew_name', '')}_{getattr(event, 'agent', '')}_{getattr(event, 'task', '')}"
        )

    def _extract_metadata(self, event: Any) -> dict:
        """Extract metadata from event for span attributes."""
        metadata = {}
        metadata["timestamp"] = event.timestamp.isoformat()

        # Add source information
        if hasattr(event, "source_type") and event.source_type:
            metadata["source_type"] = event.source_type

        if hasattr(event, "fingerprint_metadata") and event.fingerprint_metadata:
            metadata.update(event.fingerprint_metadata)

        return metadata

    # Crew event handlers
    def _handle_crew_kickoff_started(self, source: Any, event: Any) -> None:
        """Handle crew execution start."""
        run_id = self._generate_run_id(source, event)
        crew_name = event.crew_name if hasattr(event, "crew_name") else "Crew"

        metadata = self._extract_metadata(event)
        metadata["crew_name"] = crew_name

        input = getattr(event, "inputs", {})

        self._handler.start_node(
            node_type=NodeType.CHAIN.value,
            parent_run_id=None,  # Root node
            run_id=run_id,
            name=crew_name,
            input=serialize_to_str(input) if input else "-",
            metadata=metadata,
        )

    def _update_crew_input(self, run_id: str) -> None:
        """Update crew input with task descriptions. if input is not set."""
        nodes = self._handler.get_nodes()
        root_node = nodes.get(str(run_id))
        if not root_node:
            return

        input = root_node.span_params.get("input", "-")
        tasks = ""

        if not input or input == "-":
            tasks = "\n".join(
                [
                    str(node.span_params.get("metadata", {}).get("task_description"))
                    for node in nodes.values()
                    if node.span_params.get("metadata", {}).get("task_description")
                ]
            )
            input = f"Tasks: {tasks}" if tasks else "-"
            root_node.span_params["input"] = input

    def _handle_crew_kickoff_completed(self, source: Any, event: Any) -> None:
        """Handle crew execution completion."""
        run_id = self._generate_run_id(source, event)
        output = getattr(event, "output", {})
        self._update_crew_input(str(run_id))
        self._handler.end_node(run_id=run_id, output=serialize_to_str(getattr(output, "raw", output)))

    def _handle_crew_kickoff_failed(self, source: Any, event: Any) -> None:
        """Handle crew execution failure."""
        run_id = self._generate_run_id(source, event)
        metadata = self._extract_metadata(event)
        metadata["error"] = event.error

        self._handler.end_node(run_id=run_id, output=f"Error: {event.error}", metadata=metadata)

    # Agent event handlers
    def _handle_agent_execution_started(self, source: Any, event: Any) -> None:
        """Handle agent execution start."""
        run_id = self._generate_run_id(source, event)
        event_task = getattr(event, "task", None)
        parent_run_id = self._to_uuid(getattr(event_task, "id", None)) if event_task else None
        role = "Unknown Agent"
        metadata = self._extract_metadata(event)

        if hasattr(event, "agent") and event.agent:
            if hasattr(event.agent, "role"):
                role = event.agent.role
                metadata["agent_role"] = role
            if hasattr(event.agent, "id"):
                metadata["agent_id"] = str(event.agent.id)

        if hasattr(event, "tools") and event.tools:
            metadata["available_tools"] = [str(getattr(tool, "name", tool)) for tool in event.tools]

        self._handler.start_node(
            node_type=NodeType.AGENT.value,
            parent_run_id=parent_run_id,
            run_id=run_id,
            name=role,
            input=serialize_to_str(getattr(event, "task_prompt", "-")),
            metadata=metadata,
        )

    def _handle_agent_execution_completed(self, source: Any, event: Any) -> None:
        """Handle agent execution completion."""
        run_id = self._generate_run_id(source, event)
        self._handler.end_node(run_id=run_id, output=serialize_to_str(getattr(event, "output", "")))

    def _handle_agent_execution_error(self, source: Any, event: Any) -> None:
        """Handle agent execution error."""
        run_id = self._generate_run_id(source, event)
        metadata = self._extract_metadata(event)
        metadata["error"] = getattr(event, "error", "Unknown error")

        self._handler.end_node(
            run_id=run_id, output=f"Error: {getattr(event, 'error', 'Unknown error')}", metadata=metadata
        )

    # Task event handlers
    def _handle_task_started(self, source: Any, event: Any) -> None:
        """Handle task start."""
        run_id = self._generate_run_id(source, event)
        task = event.task if hasattr(event, "task") else None
        task_agent = getattr(task, "agent", None) if task else None
        task_crew = getattr(task_agent, "crew", None) if task_agent else None
        parent_run_id = self._to_uuid(getattr(task_crew, "id", None)) if task_crew else None

        metadata = self._extract_metadata(event)
        if task:
            if hasattr(task, "description"):
                metadata["task_description"] = task.description
            if hasattr(task, "id"):
                metadata["task_id"] = str(task.id)

        task_name = "Unknown Task"
        if task and hasattr(task, "description"):
            # Use first 50 chars of description as name
            task_name = task.description[:50] + "..." if len(task.description) > 50 else task.description

        input = getattr(event, "context", "")
        if not input:
            input = task.description if task and hasattr(task, "description") else ""

        self._handler.start_node(
            node_type=NodeType.CHAIN.value,
            parent_run_id=parent_run_id,
            run_id=run_id,
            name=task_name,
            input=serialize_to_str(input) if input else "-",
            metadata=metadata,
        )

    def _handle_task_completed(self, source: Any, event: Any) -> None:
        """Handle task completion."""
        run_id = self._generate_run_id(source, event)
        output = ""
        if hasattr(event, "output") and event.output:
            output = event.output.raw if hasattr(event.output, "raw") else str(event.output)

        self._handler.end_node(run_id=run_id, output=serialize_to_str(output))

    def _handle_task_failed(self, source: Any, event: Any) -> None:
        """Handle task failure."""
        run_id = self._generate_run_id(source, event)
        metadata = self._extract_metadata(event)
        metadata["error"] = event.error

        self._handler.end_node(run_id=run_id, output=f"Error: {event.error}", metadata=metadata)

    # Tool event handlers
    def _handle_tool_usage_started(self, source: Any, event: Any) -> None:
        """Handle tool usage start."""
        run_id = self._generate_run_id(source, event)
        event_agent = getattr(event, "agent", None)
        parent_run_id = self._to_uuid(getattr(event_agent, "id", None)) if event_agent else None
        # Fallback: crewAI >= 1.0 may provide agent_id as a top-level string field instead of agent.id
        if not parent_run_id:
            parent_run_id = self._to_uuid(getattr(event, "agent_id", None))
        input = getattr(event, "tool_args", {})
        tool_name = getattr(event, "tool_name", "Unknown Tool")

        metadata = self._extract_metadata(event)
        metadata["tool_name"] = tool_name
        if input:
            metadata["tool_args"] = str(input)

        self._handler.start_node(
            node_type=NodeType.TOOL.value,
            parent_run_id=parent_run_id,
            run_id=run_id,
            name=tool_name,
            input=serialize_to_str(input) if input else "-",
            metadata=metadata,
        )

    def _handle_tool_usage_finished(self, source: Any, event: Any) -> None:
        """Handle tool usage completion."""
        run_id = self._generate_run_id(source, event)
        self._handler.end_node(run_id=run_id, output=serialize_to_str(getattr(event, "output", "")))

    def _handle_tool_usage_error(self, source: Any, event: Any) -> None:
        """Handle tool usage error."""
        run_id = self._generate_run_id(source, event)
        metadata = self._extract_metadata(event)
        metadata["error"] = getattr(event, "error", "Unknown error")

        self._handler.end_node(
            run_id=run_id, output=f"Error: {getattr(event, 'error', 'Unknown error')}", metadata=metadata
        )

    def _to_uuid(self, id: Union[str, None, UUID]) -> Union[UUID, None]:
        if isinstance(id, UUID):
            return id
        if isinstance(id, str):
            try:
                return UUID(id)
            except ValueError:
                return None
        return None

    # LLM event handlers
    def _handle_llm_call_started(self, source: Any, event: Any) -> None:
        """Handle LLM call start."""
        run_id = self._generate_run_id(source, event)
        parent_run_id = self._to_uuid(getattr(event, "agent_id", None))
        llm_name = getattr(event, "model", None) or getattr(source, "model", "Unknown Model")

        metadata = self._extract_metadata(event)
        metadata["model"] = llm_name
        if hasattr(source, "temperature"):
            metadata["temperature"] = str(source.temperature)

        self._handler.start_node(
            node_type=NodeType.LLM.value,
            parent_run_id=parent_run_id,
            run_id=run_id,
            name=llm_name,
            input=serialize_to_str(getattr(event, "messages", [])),
            model=llm_name,
            temperature=getattr(source, "temperature", None),
            metadata=metadata,
        )

    def _handle_llm_call_completed(self, source: Any, event: Any) -> None:
        """Handle LLM call completion."""
        run_id = self._generate_run_id(source, event)

        token_kwargs: dict[str, Any] = {}
        response = getattr(event, "response", None)
        if response is not None:
            usage = getattr(response, "usage", None) or (getattr(response, "model_extra", None) or {}).get("usage")
            if usage is None:
                # Fallback: some crewai/litellm versions expose usage on the source object
                usage = getattr(source, "_token_usage", None)
            if usage is not None:
                _get = usage.get if isinstance(usage, dict) else lambda k, d=0: getattr(usage, k, d)
                token_kwargs["num_input_tokens"] = _get("prompt_tokens", 0)
                token_kwargs["num_output_tokens"] = _get("completion_tokens", 0)
                token_kwargs["total_tokens"] = _get("total_tokens", 0)

        self._handler.end_node(run_id=run_id, output=serialize_to_str(response or ""), **token_kwargs)

    def _handle_llm_call_failed(self, source: Any, event: Any) -> None:
        """Handle LLM call failure."""
        run_id = self._generate_run_id(source, event)
        metadata = self._extract_metadata(event)
        metadata["error"] = getattr(event, "error", "Unknown error")

        self._handler.end_node(
            run_id=run_id, output=f"Error: {getattr(event, 'error', 'Unknown error')}", metadata=metadata
        )

    # Memory event handlers
    def _handle_memory_query_started(self, source: Any, event: Any) -> None:
        """Handle memory query start."""
        run_id = self._generate_run_id(source, event)
        parent_run_id = self._to_uuid(getattr(event, "agent_id", None))
        parent_node = self._handler.get_node(parent_run_id) if parent_run_id else None

        if not parent_node and parent_run_id:
            task_id = getattr(event, "task_id", None)
            if task_id:
                task_run_id = self._to_uuid(task_id)
                parent_node = self._handler.get_node(task_run_id) if task_run_id else None
                if parent_node:
                    self._handler.start_node(
                        node_type=NodeType.AGENT.value,
                        parent_run_id=task_run_id,
                        run_id=parent_run_id,
                        name=getattr(event, "agent_role", "Unknown Agent"),
                        input=serialize_to_str(event.query),
                    )

        metadata = self._extract_metadata(event)
        metadata["query"] = event.query
        metadata["limit"] = event.limit
        if event.score_threshold is not None:
            metadata["score_threshold"] = event.score_threshold
        if hasattr(event, "agent_role") and event.agent_role:
            metadata["agent_role"] = event.agent_role

        self._handler.start_node(
            node_type=NodeType.RETRIEVER.value,
            parent_run_id=parent_run_id,
            run_id=run_id,
            name="Memory Query",
            input=serialize_to_str(event.query),
            metadata=metadata,
        )

    def _handle_memory_query_completed(self, source: Any, event: Any) -> None:
        """Handle memory query completion."""
        run_id = self._generate_run_id(source, event)

        metadata = self._extract_metadata(event)
        metadata["query_time_ms"] = event.query_time_ms
        metadata["results_count"] = len(event.results) if hasattr(event.results, "__len__") else 0

        output = serialize_to_str(event.results) if event.results else "No results found"
        self._handler.end_node(run_id=run_id, output=output, metadata=metadata)

    def _handle_memory_query_failed(self, source: Any, event: Any) -> None:
        """Handle memory query failure."""
        run_id = self._generate_run_id(source, event)

        metadata = self._extract_metadata(event)
        metadata["error"] = event.error

        self._handler.end_node(run_id=run_id, output=f"Memory query failed: {event.error}", metadata=metadata)

    def _handle_memory_save_started(self, source: Any, event: Any) -> None:
        """Handle memory save start."""
        run_id = self._generate_run_id(source, event)
        parent_run_id = self._to_uuid(getattr(event, "agent_id", None))

        metadata = self._extract_metadata(event)
        if hasattr(event, "metadata") and event.metadata:
            metadata.update(event.metadata)
        if hasattr(event, "agent_role") and event.agent_role:
            metadata["agent_role"] = event.agent_role

        input_value = event.value if event.value else "Memory content"

        self._handler.start_node(
            node_type=NodeType.CHAIN.value,
            parent_run_id=parent_run_id,
            run_id=run_id,
            name="Memory Save",
            input=serialize_to_str(input_value),
            metadata=metadata,
        )

    def _handle_memory_save_completed(self, source: Any, event: Any) -> None:
        """Handle memory save completion."""
        run_id = self._generate_run_id(source, event)

        metadata = self._extract_metadata(event)
        metadata["save_time_ms"] = event.save_time_ms

        self._handler.end_node(run_id=run_id, output=f"Memory saved successfully: {event.value}", metadata=metadata)

    def _handle_memory_save_failed(self, source: Any, event: Any) -> None:
        """Handle memory save failure."""
        run_id = self._generate_run_id(source, event)

        metadata = self._extract_metadata(event)
        metadata["error"] = event.error

        self._handler.end_node(run_id=run_id, output=f"Memory save failed: {event.error}", metadata=metadata)

    def _handle_memory_retrieval_started(self, source: Any, event: Any) -> None:
        """Handle memory retrieval start."""
        run_id = self._generate_run_id(source, event)
        parent_run_id = self._to_uuid(getattr(event, "task_id", None))

        metadata = self._extract_metadata(event)
        if hasattr(event, "task_id") and event.task_id:
            metadata["task_id"] = event.task_id

        self._handler.start_node(
            node_type=NodeType.CHAIN.value,
            parent_run_id=parent_run_id,
            run_id=run_id,
            name="Memory Retrieval",
            input="Retrieving relevant memories for task",
            metadata=metadata,
        )

    def _handle_memory_retrieval_completed(self, source: Any, event: Any) -> None:
        """Handle memory retrieval completion."""
        run_id = self._generate_run_id(source, event)

        metadata = self._extract_metadata(event)
        metadata["retrieval_time_ms"] = event.retrieval_time_ms

        self._handler.end_node(run_id=run_id, output=serialize_to_str(event.memory_content), metadata=metadata)

    def _handle_memory_retrieval_failed(self, source: Any, event: Any) -> None:
        """Handle memory retrieval failure."""
        run_id = self._generate_run_id(source, event)

        metadata = self._extract_metadata(event)
        metadata["error"] = getattr(event, "error", "Unknown error")

        self._handler.end_node(run_id=run_id, output=f"Memory retrieval failed: {metadata['error']}", metadata=metadata)

    def lite_llm_usage_callback(
        self,
        kwargs: dict,  # kwargs to completion
        completion_response: Any,  # response from completion
        start_time: datetime,
        end_time: datetime,
    ) -> None:
        node_id = self._generate_run_id(kwargs, kwargs)

        node = self._handler.get_node(node_id)
        if not node:
            _logger.debug(f"No node exists for run_id {node_id}")
            return
        usage = completion_response.model_extra["usage"]
        node.span_params["usage"] = usage.model_dump() if hasattr(usage, "model_dump") else usage
        node.span_params["num_input_tokens"] = getattr(usage, "prompt_tokens", 0)
        node.span_params["num_output_tokens"] = getattr(usage, "completion_tokens", 0)
        node.span_params["total_tokens"] = getattr(usage, "total_tokens", 0)



from collections.abc import Mapping
from typing import Any, Literal, TypeVar, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.metric_color import MetricColor

T = TypeVar("T", bound="BooleanColorConstraint")


@_attrs_define
class BooleanColorConstraint:
    """A color constraint for boolean metric values.

    Assigns a color when a boolean score equals the given value.

    Only the 'eq' operator is supported.

    Example:
        {"color": "green", "operator": "eq", "value": true}
        {"color": "red", "operator": "eq", "value": false}

    Attributes
    ----------
            color (MetricColor): Allowed colors for metric threshold visualization in the UI.
            operator (Literal['eq']):
            value (bool):
    """

    color: MetricColor
    operator: Literal["eq"]
    value: bool
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        color = self.color.value

        operator = self.operator

        value = self.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"color": color, "operator": operator, "value": value})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        color = MetricColor(d.pop("color"))

        operator = cast(Literal["eq"], d.pop("operator"))
        if operator != "eq":
            raise ValueError(f"operator must match const 'eq', got '{operator}'")

        value = d.pop("value")

        boolean_color_constraint = cls(color=color, operator=operator, value=value)

        boolean_color_constraint.additional_properties = d
        return boolean_color_constraint

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Deprecated: use galileo.project instead of galileo.__future__.project."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.project is deprecated. Use galileo.project instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.project import Project  # noqa: E402

__all__ = ["Project"]



"""Deprecated: use the galileo top-level package instead of galileo.__future__."""

from __future__ import annotations

import warnings

warnings.warn(
    "Importing from galileo.__future__ is deprecated. Use the galileo top-level package instead "
    "(e.g., 'from galileo import Project').",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.collaborator import Collaborator, CollaboratorRole  # noqa: E402
from galileo.configuration import Configuration  # noqa: E402
from galileo.dataset import Dataset  # noqa: E402
from galileo.experiment import Experiment  # noqa: E402
from galileo.integration import Integration  # noqa: E402
from galileo.log_stream import LogStream  # noqa: E402
from galileo.metric import CodeMetric, GalileoMetric, LlmMetric, LocalMetric, Metric  # noqa: E402
from galileo.model import Model  # noqa: E402
from galileo.project import Project  # noqa: E402
from galileo.prompt import Prompt  # noqa: E402
from galileo.schema.message import Message  # noqa: E402
from galileo.search import RecordType  # noqa: E402
from galileo.shared.exceptions import (  # noqa: E402
    APIError,
    ConfigurationError,
    GalileoFutureError,
    ResourceConflictError,
    ResourceNotFoundError,
    ValidationError,
)
from galileo.utils.log_config import enable_console_logging  # noqa: E402
from galileo_core.schemas.logging.llm import MessageRole  # noqa: E402
from galileo_core.schemas.logging.step import StepType  # noqa: E402

__all__ = [
    "APIError",
    "CodeMetric",
    "Collaborator",
    "CollaboratorRole",
    "Configuration",
    "ConfigurationError",
    "Dataset",
    "Experiment",
    "GalileoFutureError",
    "GalileoMetric",
    "Integration",
    "LlmMetric",
    "LocalMetric",
    "LogStream",
    "Message",
    "MessageRole",
    "Metric",
    "Model",
    "Project",
    "Prompt",
    "RecordType",
    "ResourceConflictError",
    "ResourceNotFoundError",
    "StepType",
    "ValidationError",
    "enable_console_logging",
]



"""Galileo."""

from galileo.collaborator import Collaborator, CollaboratorRole
from galileo.configuration import Configuration
from galileo.dataset import Dataset
from galileo.decorator import GalileoDecorator, galileo_context, log, start_session
from galileo.exceptions import (
    AuthenticationError,
    BadRequestError,
    ConflictError,
    ForbiddenError,
    GalileoAPIError,
    GalileoLoggerException,
    NotFoundError,
    RateLimitError,
    ServerError,
)
from galileo.experiment import Experiment
from galileo.integration import Integration
from galileo.log_stream import LogStream
from galileo.logger import GalileoLogger
from galileo.metric import CodeMetric, GalileoMetric, LlmMetric, LocalMetric, Metric
from galileo.model import Model
from galileo.project import Project
from galileo.prompt import Prompt
from galileo.protect import ainvoke_protect, invoke_protect
from galileo.provider import AnthropicProvider, AzureProvider, BedrockProvider, OpenAIProvider, Provider
from galileo.schema.message import Message
from galileo.schema.metrics import GalileoMetrics, GalileoScorers
from galileo.shared.base import SyncState
from galileo.shared.exceptions import (
    APIError,
    ConfigurationError,
    GalileoFutureError,
    ResourceConflictError,
    ResourceNotFoundError,
    ValidationError,
)
from galileo.stages import (
    create_protect_stage,
    get_protect_stage,
    pause_protect_stage,
    resume_protect_stage,
    update_protect_stage,
)
from galileo.tracing import get_tracing_headers
from galileo.types import MetricSpec
from galileo.utils.log_config import enable_console_logging
from galileo_core.helpers.api_key import create_api_key, delete_api_key, list_api_keys
from galileo_core.helpers.dependencies import is_dependency_available
from galileo_core.schemas.logging.llm import MessageRole, ToolCall, ToolCallFunction
from galileo_core.schemas.logging.session import Session
from galileo_core.schemas.logging.span import (
    AgentSpan,
    LlmSpan,
    RetrieverSpan,
    Span,
    StepWithChildSpans,
    ToolSpan,
    WorkflowSpan,
)
from galileo_core.schemas.logging.step import StepType
from galileo_core.schemas.logging.trace import Trace
from galileo_core.schemas.protect.execution_status import ExecutionStatus
from galileo_core.schemas.protect.payload import Payload
from galileo_core.schemas.protect.request import Request
from galileo_core.schemas.protect.response import Response
from galileo_core.schemas.protect.ruleset import Ruleset
from galileo_core.schemas.protect.stage import StageType

__version__ = "2.0.0"

__all__ = [
    "APIError",
    "AgentSpan",
    "AnthropicProvider",
    "AuthenticationError",
    "AzureProvider",
    "BadRequestError",
    "BedrockProvider",
    "CodeMetric",
    "Collaborator",
    "CollaboratorRole",
    "Configuration",
    "ConfigurationError",
    "ConflictError",
    "Dataset",
    "ExecutionStatus",
    "Experiment",
    "ForbiddenError",
    "GalileoAPIError",
    "GalileoDecorator",
    "GalileoFutureError",
    "GalileoLogger",
    "GalileoLoggerException",
    "GalileoMetric",
    "GalileoMetrics",
    "GalileoScorers",
    "Integration",
    "LlmMetric",
    "LlmSpan",
    "LocalMetric",
    "LogStream",
    "Message",
    "MessageRole",
    "Metric",
    "MetricSpec",
    "Model",
    "NotFoundError",
    "OpenAIProvider",
    "Payload",
    "Project",
    "Prompt",
    "Provider",
    "RateLimitError",
    "Request",
    "ResourceConflictError",
    "ResourceNotFoundError",
    "Response",
    "RetrieverSpan",
    "Ruleset",
    "ServerError",
    "Session",
    "Span",
    "StageType",
    "StepType",
    "StepWithChildSpans",
    "SyncState",
    "ToolCall",
    "ToolCallFunction",
    "ToolSpan",
    "Trace",
    "ValidationError",
    "WorkflowSpan",
    "ainvoke_protect",
    "create_api_key",
    "create_protect_stage",
    "delete_api_key",
    "enable_console_logging",
    "galileo_context",
    "get_protect_stage",
    "get_tracing_headers",
    "invoke_protect",
    "is_dependency_available",
    "list_api_keys",
    "log",
    "pause_protect_stage",
    "resume_protect_stage",
    "start_session",
    "update_protect_stage",
]



"""Deprecated: use galileo.collaborator instead of galileo.__future__.collaborator."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.collaborator is deprecated. Use galileo.collaborator instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.collaborator import Collaborator, CollaboratorRole  # noqa: E402

__all__ = ["Collaborator", "CollaboratorRole"]



"""Deprecated: use galileo.configuration instead of galileo.__future__.configuration."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.configuration is deprecated. Use galileo.configuration instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.configuration import (  # noqa: E402
    _CONFIGURATION_KEYS,
    _KEYS_BY_NAME,
    VALID_LOG_LEVELS,
    ConfigKey,
    Configuration,
    ConfigurationMeta,
    parse_log_level,
)

__all__ = [
    "VALID_LOG_LEVELS",
    "_CONFIGURATION_KEYS",
    "_KEYS_BY_NAME",
    "ConfigKey",
    "Configuration",
    "ConfigurationMeta",
    "parse_log_level",
]



from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.metric_color import MetricColor
from ..models.numeric_color_constraint_operator import NumericColorConstraintOperator

T = TypeVar("T", bound="NumericColorConstraint")


@_attrs_define
class NumericColorConstraint:
    """A color constraint for numeric metric values.

    Assigns a color when a numeric score matches the given operator and value.

    Operators and expected value shapes:
      - eq, gt, gte, lt, lte: value must be a single float.
      - between: value must be a list of exactly 2 floats [low, high] where low < high.
        The range is inclusive on both ends.

    Example:
        {"color": "green", "operator": "gte", "value": 0.8}
        {"color": "yellow", "operator": "between", "value": [0.3, 0.7]}

    Attributes
    ----------
            color (MetricColor): Allowed colors for metric threshold visualization in the UI.
            operator (NumericColorConstraintOperator):
            value (Union[float, list[float]]):
    """

    color: MetricColor
    operator: NumericColorConstraintOperator
    value: Union[float, list[float]]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        color = self.color.value

        operator = self.operator.value

        value: Union[float, list[float]]
        value = self.value if isinstance(self.value, list) else self.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"color": color, "operator": operator, "value": value})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        color = MetricColor(d.pop("color"))

        operator = NumericColorConstraintOperator(d.pop("operator"))

        def _parse_value(data: object) -> Union[float, list[float]]:
            try:
                if not isinstance(data, list):
                    raise TypeError()
                return cast(list[float], data)

            except:  # noqa: E722
                pass
            return cast(Union[float, list[float]], data)

        value = _parse_value(d.pop("value"))

        numeric_color_constraint = cls(color=color, operator=operator, value=value)

        numeric_color_constraint.additional_properties = d
        return numeric_color_constraint

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.categorical_color_constraint import CategoricalColorConstraint


T = TypeVar("T", bound="MetricColorPickerMultiLabel")


@_attrs_define
class MetricColorPickerMultiLabel:
    """Color picker configuration for multi-label metrics.

    Behaves the same as categorical but intended for metrics that produce
    multiple labels. A category value must not appear in more than one constraint.

    Example:
        {
            "type": "multi_label",
            "constraints": [
                {"color": "green", "operator": "eq", "value": "relevant"},
                {"color": "yellow", "operator": "one_of", "value": ["partial", "related"]}
            ]
        }

    Attributes
    ----------
            constraints (list['CategoricalColorConstraint']):
            type_ (Union[Literal['multi_label'], Unset]):  Default: 'multi_label'.
    """

    constraints: list["CategoricalColorConstraint"]
    type_: Union[Literal["multi_label"], Unset] = "multi_label"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        constraints = []
        for constraints_item_data in self.constraints:
            constraints_item = constraints_item_data.to_dict()
            constraints.append(constraints_item)

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"constraints": constraints})
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.categorical_color_constraint import CategoricalColorConstraint

        d = dict(src_dict)
        constraints = []
        _constraints = d.pop("constraints")
        for constraints_item_data in _constraints:
            constraints_item = CategoricalColorConstraint.from_dict(constraints_item_data)

            constraints.append(constraints_item)

        type_ = cast(Union[Literal["multi_label"], Unset], d.pop("type", UNSET))
        if type_ != "multi_label" and not isinstance(type_, Unset):
            raise ValueError(f"type must match const 'multi_label', got '{type_}'")

        metric_color_picker_multi_label = cls(constraints=constraints, type_=type_)

        metric_color_picker_multi_label.additional_properties = d
        return metric_color_picker_multi_label

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="ScoreBucket")


@_attrs_define
class ScoreBucket:
    """
    Attributes
    ----------
        min_inclusive (int):
        max_exclusive (Union[None, int]):
        count (int):
    """

    min_inclusive: int
    max_exclusive: Union[None, int]
    count: int
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        min_inclusive = self.min_inclusive

        max_exclusive: Union[None, int]
        max_exclusive = self.max_exclusive

        count = self.count

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"min_inclusive": min_inclusive, "max_exclusive": max_exclusive, "count": count})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        min_inclusive = d.pop("min_inclusive")

        def _parse_max_exclusive(data: object) -> Union[None, int]:
            if data is None:
                return data
            return cast(Union[None, int], data)

        max_exclusive = _parse_max_exclusive(d.pop("max_exclusive"))

        count = d.pop("count")

        score_bucket = cls(min_inclusive=min_inclusive, max_exclusive=max_exclusive, count=count)

        score_bucket.additional_properties = d
        return score_bucket

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Deprecated: use galileo.shared.column instead of galileo.__future__.shared.column."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.shared.column is deprecated. Use galileo.shared.column instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.shared.column import Column, ColumnCollection, _unwrap_unset  # noqa: E402

__all__ = ["Column", "ColumnCollection", "_unwrap_unset"]



"""Deprecated: use galileo.dataset instead of galileo.__future__.dataset."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.dataset is deprecated. Use galileo.dataset instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.dataset import Dataset  # noqa: E402

__all__ = ["Dataset"]



"""Deprecated: use galileo.types instead of galileo.__future__.types."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.types is deprecated. Use galileo.types instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.types import MetricSpec  # noqa: E402

__all__ = ["MetricSpec"]



"""Deprecated: use galileo.model instead of galileo.__future__.model."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.model is deprecated. Use galileo.model instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.model import Model  # noqa: E402

__all__ = ["Model"]



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.numeric_color_constraint import NumericColorConstraint


T = TypeVar("T", bound="MetricColorPickerNumeric")


@_attrs_define
class MetricColorPickerNumeric:
    """Color picker configuration for numeric metrics.

    Each constraint maps a numeric condition to a color. The UI uses these
    constraints to color-code metric values (e.g. green for high scores,
    red for low scores).

    Example:
        {
            "type": "numeric",
            "constraints": [
                {"color": "green", "operator": "gte", "value": 0.8},
                {"color": "yellow", "operator": "between", "value": [0.3, 0.8]},
                {"color": "red", "operator": "lt", "value": 0.3}
            ]
        }

    Attributes
    ----------
            constraints (list['NumericColorConstraint']):
            type_ (Union[Literal['numeric'], Unset]):  Default: 'numeric'.
    """

    constraints: list["NumericColorConstraint"]
    type_: Union[Literal["numeric"], Unset] = "numeric"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        constraints = []
        for constraints_item_data in self.constraints:
            constraints_item = constraints_item_data.to_dict()
            constraints.append(constraints_item)

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"constraints": constraints})
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.numeric_color_constraint import NumericColorConstraint

        d = dict(src_dict)
        constraints = []
        _constraints = d.pop("constraints")
        for constraints_item_data in _constraints:
            constraints_item = NumericColorConstraint.from_dict(constraints_item_data)

            constraints.append(constraints_item)

        type_ = cast(Union[Literal["numeric"], Unset], d.pop("type", UNSET))
        if type_ != "numeric" and not isinstance(type_, Unset):
            raise ValueError(f"type must match const 'numeric', got '{type_}'")

        metric_color_picker_numeric = cls(constraints=constraints, type_=type_)

        metric_color_picker_numeric.additional_properties = d
        return metric_color_picker_numeric

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.and_node_log_records_filter import AndNodeLogRecordsFilter
    from ..models.filter_leaf_log_records_filter import FilterLeafLogRecordsFilter
    from ..models.metric_aggregation_detail import MetricAggregationDetail
    from ..models.not_node_log_records_filter import NotNodeLogRecordsFilter
    from ..models.or_node_log_records_filter import OrNodeLogRecordsFilter


T = TypeVar("T", bound="LogRecordsCustomMetricsQueryRequest")


@_attrs_define
class LogRecordsCustomMetricsQueryRequest:
    """
    Attributes
    ----------
        start_time (datetime.datetime): Include traces from this time onward.
        end_time (datetime.datetime): Include traces up to this time.
        metric_details (list['MetricAggregationDetail']): List of metrics to aggregate with their widget IDs and
            aggregation types (max 100)
        log_stream_id (Union[None, Unset, str]): Log stream id associated with the traces.
        experiment_id (Union[None, Unset, str]): Experiment id associated with the traces.
        metrics_testing_id (Union[None, Unset, str]): Metrics testing id associated with the traces.
        filter_tree (Union['AndNodeLogRecordsFilter', 'FilterLeafLogRecordsFilter', 'NotNodeLogRecordsFilter',
            'OrNodeLogRecordsFilter', None, Unset]): Filter expression tree for complex filtering
        interval_minutes (Union[Unset, int]): Time interval in minutes for bucketing Default: 5.
        group_by (Union[None, Unset, str]): Column to group by.
    """

    start_time: datetime.datetime
    end_time: datetime.datetime
    metric_details: list["MetricAggregationDetail"]
    log_stream_id: Union[None, Unset, str] = UNSET
    experiment_id: Union[None, Unset, str] = UNSET
    metrics_testing_id: Union[None, Unset, str] = UNSET
    filter_tree: Union[
        "AndNodeLogRecordsFilter",
        "FilterLeafLogRecordsFilter",
        "NotNodeLogRecordsFilter",
        "OrNodeLogRecordsFilter",
        None,
        Unset,
    ] = UNSET
    interval_minutes: Union[Unset, int] = 5
    group_by: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.and_node_log_records_filter import AndNodeLogRecordsFilter
        from ..models.filter_leaf_log_records_filter import FilterLeafLogRecordsFilter
        from ..models.not_node_log_records_filter import NotNodeLogRecordsFilter
        from ..models.or_node_log_records_filter import OrNodeLogRecordsFilter

        start_time = self.start_time.isoformat()

        end_time = self.end_time.isoformat()

        metric_details = []
        for metric_details_item_data in self.metric_details:
            metric_details_item = metric_details_item_data.to_dict()
            metric_details.append(metric_details_item)

        log_stream_id: Union[None, Unset, str]
        log_stream_id = UNSET if isinstance(self.log_stream_id, Unset) else self.log_stream_id

        experiment_id: Union[None, Unset, str]
        experiment_id = UNSET if isinstance(self.experiment_id, Unset) else self.experiment_id

        metrics_testing_id: Union[None, Unset, str]
        metrics_testing_id = UNSET if isinstance(self.metrics_testing_id, Unset) else self.metrics_testing_id

        filter_tree: Union[None, Unset, dict[str, Any]]
        if isinstance(self.filter_tree, Unset):
            filter_tree = UNSET
        elif isinstance(
            self.filter_tree,
            (FilterLeafLogRecordsFilter, AndNodeLogRecordsFilter, OrNodeLogRecordsFilter, NotNodeLogRecordsFilter),
        ):
            filter_tree = self.filter_tree.to_dict()
        else:
            filter_tree = self.filter_tree

        interval_minutes = self.interval_minutes

        group_by: Union[None, Unset, str]
        group_by = UNSET if isinstance(self.group_by, Unset) else self.group_by

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"start_time": start_time, "end_time": end_time, "metric_details": metric_details})
        if log_stream_id is not UNSET:
            field_dict["log_stream_id"] = log_stream_id
        if experiment_id is not UNSET:
            field_dict["experiment_id"] = experiment_id
        if metrics_testing_id is not UNSET:
            field_dict["metrics_testing_id"] = metrics_testing_id
        if filter_tree is not UNSET:
            field_dict["filter_tree"] = filter_tree
        if interval_minutes is not UNSET:
            field_dict["interval_minutes"] = interval_minutes
        if group_by is not UNSET:
            field_dict["group_by"] = group_by

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.and_node_log_records_filter import AndNodeLogRecordsFilter
        from ..models.filter_leaf_log_records_filter import FilterLeafLogRecordsFilter
        from ..models.metric_aggregation_detail import MetricAggregationDetail
        from ..models.not_node_log_records_filter import NotNodeLogRecordsFilter
        from ..models.or_node_log_records_filter import OrNodeLogRecordsFilter

        d = dict(src_dict)
        start_time = isoparse(d.pop("start_time"))

        end_time = isoparse(d.pop("end_time"))

        metric_details = []
        _metric_details = d.pop("metric_details")
        for metric_details_item_data in _metric_details:
            metric_details_item = MetricAggregationDetail.from_dict(metric_details_item_data)

            metric_details.append(metric_details_item)

        def _parse_log_stream_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        log_stream_id = _parse_log_stream_id(d.pop("log_stream_id", UNSET))

        def _parse_experiment_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        experiment_id = _parse_experiment_id(d.pop("experiment_id", UNSET))

        def _parse_metrics_testing_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        metrics_testing_id = _parse_metrics_testing_id(d.pop("metrics_testing_id", UNSET))

        def _parse_filter_tree(
            data: object,
        ) -> Union[
            "AndNodeLogRecordsFilter",
            "FilterLeafLogRecordsFilter",
            "NotNodeLogRecordsFilter",
            "OrNodeLogRecordsFilter",
            None,
            Unset,
        ]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return FilterLeafLogRecordsFilter.from_dict(data)

            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return AndNodeLogRecordsFilter.from_dict(data)

            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return OrNodeLogRecordsFilter.from_dict(data)

            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return NotNodeLogRecordsFilter.from_dict(data)

            except:  # noqa: E722
                pass
            return cast(
                Union[
                    "AndNodeLogRecordsFilter",
                    "FilterLeafLogRecordsFilter",
                    "NotNodeLogRecordsFilter",
                    "OrNodeLogRecordsFilter",
                    None,
                    Unset,
                ],
                data,
            )

        filter_tree = _parse_filter_tree(d.pop("filter_tree", UNSET))

        interval_minutes = d.pop("interval_minutes", UNSET)

        def _parse_group_by(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        group_by = _parse_group_by(d.pop("group_by", UNSET))

        log_records_custom_metrics_query_request = cls(
            start_time=start_time,
            end_time=end_time,
            metric_details=metric_details,
            log_stream_id=log_stream_id,
            experiment_id=experiment_id,
            metrics_testing_id=metrics_testing_id,
            filter_tree=filter_tree,
            interval_minutes=interval_minutes,
            group_by=group_by,
        )

        log_records_custom_metrics_query_request.additional_properties = d
        return log_records_custom_metrics_query_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ErrorType(str, Enum):
    CONFIGURATION_ERROR = "configuration_error"
    CREDENTIALS_ERROR = "credentials_error"
    DATA_VALIDATION_ERROR = "data_validation_error"
    LLM_API_ERROR = "llm_api_error"
    NOT_APPLICABLE_REASON = "not_applicable_reason"
    NOT_FOUND_ERROR = "not_found_error"
    PERMISSION_ERROR = "permission_error"
    SYSTEM_ERROR = "system_error"
    UNCATALOGED_ERROR = "uncataloged_error"
    WORKFLOW_ERROR = "workflow_error"

    def __str__(self) -> str:
        return str(self.value)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.log_records_metrics_response_aggregate_metrics import LogRecordsMetricsResponseAggregateMetrics
    from ..models.log_records_metrics_response_bucketed_metrics import LogRecordsMetricsResponseBucketedMetrics
    from ..models.log_records_metrics_response_standard_errors_type_0 import (
        LogRecordsMetricsResponseStandardErrorsType0,
    )


T = TypeVar("T", bound="LogRecordsMetricsResponse")


@_attrs_define
class LogRecordsMetricsResponse:
    """
    Attributes
    ----------
        group_by_columns (list[str]):
        aggregate_metrics (LogRecordsMetricsResponseAggregateMetrics):
        bucketed_metrics (LogRecordsMetricsResponseBucketedMetrics):
        ems_captured_error (Union[Unset, bool]): Whether any EMS error codes were encountered in the queried metrics
            Default: False.
        standard_errors (Union['LogRecordsMetricsResponseStandardErrorsType0', None, Unset]): Structured EMS errors for
            each error code encountered, keyed by code.
    """

    group_by_columns: list[str]
    aggregate_metrics: "LogRecordsMetricsResponseAggregateMetrics"
    bucketed_metrics: "LogRecordsMetricsResponseBucketedMetrics"
    ems_captured_error: Union[Unset, bool] = False
    standard_errors: Union["LogRecordsMetricsResponseStandardErrorsType0", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.log_records_metrics_response_standard_errors_type_0 import (
            LogRecordsMetricsResponseStandardErrorsType0,
        )

        group_by_columns = self.group_by_columns

        aggregate_metrics = self.aggregate_metrics.to_dict()

        bucketed_metrics = self.bucketed_metrics.to_dict()

        ems_captured_error = self.ems_captured_error

        standard_errors: Union[None, Unset, dict[str, Any]]
        if isinstance(self.standard_errors, Unset):
            standard_errors = UNSET
        elif isinstance(self.standard_errors, LogRecordsMetricsResponseStandardErrorsType0):
            standard_errors = self.standard_errors.to_dict()
        else:
            standard_errors = self.standard_errors

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "group_by_columns": group_by_columns,
                "aggregate_metrics": aggregate_metrics,
                "bucketed_metrics": bucketed_metrics,
            }
        )
        if ems_captured_error is not UNSET:
            field_dict["ems_captured_error"] = ems_captured_error
        if standard_errors is not UNSET:
            field_dict["standard_errors"] = standard_errors

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.log_records_metrics_response_aggregate_metrics import LogRecordsMetricsResponseAggregateMetrics
        from ..models.log_records_metrics_response_bucketed_metrics import LogRecordsMetricsResponseBucketedMetrics
        from ..models.log_records_metrics_response_standard_errors_type_0 import (
            LogRecordsMetricsResponseStandardErrorsType0,
        )

        d = dict(src_dict)
        group_by_columns = cast(list[str], d.pop("group_by_columns"))

        aggregate_metrics = LogRecordsMetricsResponseAggregateMetrics.from_dict(d.pop("aggregate_metrics"))

        bucketed_metrics = LogRecordsMetricsResponseBucketedMetrics.from_dict(d.pop("bucketed_metrics"))

        ems_captured_error = d.pop("ems_captured_error", UNSET)

        def _parse_standard_errors(data: object) -> Union["LogRecordsMetricsResponseStandardErrorsType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return LogRecordsMetricsResponseStandardErrorsType0.from_dict(data)

            except:  # noqa: E722
                pass
            return cast(Union["LogRecordsMetricsResponseStandardErrorsType0", None, Unset], data)

        standard_errors = _parse_standard_errors(d.pop("standard_errors", UNSET))

        log_records_metrics_response = cls(
            group_by_columns=group_by_columns,
            aggregate_metrics=aggregate_metrics,
            bucketed_metrics=bucketed_metrics,
            ems_captured_error=ems_captured_error,
            standard_errors=standard_errors,
        )

        log_records_metrics_response.additional_properties = d
        return log_records_metrics_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Deprecated: use galileo.log_stream instead of galileo.__future__.log_stream."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.log_stream is deprecated. Use galileo.log_stream instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.log_stream import LogStream  # noqa: E402

__all__ = ["LogStream"]



from enum import Enum


class MetricAggregation(str, Enum):
    AVERAGE = "Average"
    COUNT = "Count"
    MAX = "Max"
    MIN = "Min"
    P50 = "P50"
    P90 = "P90"
    P95 = "P95"
    P99 = "P99"
    SUM = "Sum"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class RecommendedModelPurpose(str, Enum):
    AUTOTUNE = "autotune"
    CUSTOM_METRIC_AUTOGEN = "custom_metric_autogen"
    CUSTOM_METRIC_JUDGE = "custom_metric_judge"
    SIGNALS = "signals"

    def __str__(self) -> str:
        return str(self.value)



import builtins
from typing import Optional, Union
from uuid import UUID

from galileo.resources.models.scorer_config import ScorerConfig
from galileo.resources.models.scorer_response import ScorerResponse
from galileo.schema.metrics import GalileoMetrics, LocalMetricConfig, Metric
from galileo.scorers import Scorers, ScorerSettings
from galileo_core.schemas.logging.span import Span, StepWithChildSpans
from galileo_core.schemas.logging.trace import Trace
from galileo_core.schemas.shared.metric import MetricValueType


def populate_local_metrics(step: Union[Trace, Span], local_metrics: list[LocalMetricConfig]) -> None:
    for local_metric in local_metrics:
        _populate_local_metric(step, local_metric, [])


def _populate_local_metric(
    step: Union[Trace, Span], local_metric: LocalMetricConfig, scores: list[MetricValueType]
) -> None:
    if isinstance(step, StepWithChildSpans):
        for span in step.spans:
            _populate_local_metric(span, local_metric, scores)
        if local_metric.aggregator_fn and scores and step.type in local_metric.aggregatable_types:
            aggregate_metric_result = local_metric.aggregator_fn(scores)
            if isinstance(aggregate_metric_result, dict):
                for suffix, value in aggregate_metric_result.items():
                    setattr(step.metrics, local_metric.name + "_" + suffix.removeprefix("_"), value)
            else:
                setattr(step.metrics, local_metric.name, aggregate_metric_result)
    if step.type in local_metric.scorable_types:
        metric_value = local_metric.scorer_fn(step)
        setattr(step.metrics, local_metric.name, local_metric.scorer_fn(step))
        scores.append(metric_value)


def _is_uuid(value: str) -> bool:
    """Check if a string is a valid UUID."""
    try:
        UUID(value)
        return True
    except ValueError:
        return False


def create_metric_configs(
    project_id: str,
    run_id: Optional[str],  # Can be experiment_id, log_stream_id, or None (for trigger=True flow)
    metrics: builtins.list[Union[GalileoMetrics, Metric, LocalMetricConfig, str]],
) -> tuple[builtins.list[ScorerConfig], builtins.list[LocalMetricConfig]]:
    """
    Process metrics and create scorer configurations for experiments or log streams.

    This unified function categorizes metrics into server-side and client-side types,
    validates they exist, and registers server-side metrics with Galileo.

    Metrics can be specified as:
    - GalileoMetrics enum values (human-readable labels like "Correctness")
    - Metric objects with name and optional version
    - UUID strings (scorer IDs for direct lookup)
    - Plain strings (searched by label with name fallback)
    - LocalMetricConfig objects (processed client-side)

    Parameters
    ----------
    project_id : str
        The ID of the project
    run_id : Optional[str]
        The ID of the run (can be experiment ID or log stream ID).
        When None, scorer registration is skipped (trigger=True flow).
    metrics : list
        List of metrics to configure

    Returns
    -------
    tuple[list[ScorerConfig], list[LocalMetricConfig]]
        A tuple containing:
        - List of ScorerConfig objects for server-side metrics configured in Galileo
        - List of LocalMetricConfig objects for client-side metrics to process locally

    Raises
    ------
    ValueError
        If any specified metrics are unknown or don't exist in Galileo
    """
    local_metric_configs: list[LocalMetricConfig] = []
    scorer_ids: list[str] = []
    label_searches: list[tuple[str, Optional[int]]] = []

    # Categorize metrics by type
    for metric in metrics:
        if isinstance(metric, GalileoMetrics):
            label_searches.append((metric.value, None))
        elif isinstance(metric, Metric):
            label_searches.append((metric.name, metric.version))
        elif isinstance(metric, LocalMetricConfig):
            local_metric_configs.append(metric)
        elif isinstance(metric, str):
            if _is_uuid(metric):
                scorer_ids.append(metric)
            else:
                label_searches.append((metric, None))
        else:
            raise ValueError(f"Unknown metric type: {type(metric)}")

    # Process server-side metrics with targeted API calls
    scorers_client = Scorers()
    all_resolved: list[ScorerResponse] = []

    if scorer_ids:
        all_resolved.extend(scorers_client.list_by_ids(scorer_ids))

    if label_searches:
        search_values = list({v for v, _ in label_searches})
        all_resolved.extend(scorers_client.list_by_labels(search_values, strict=False))

    # Build lookup maps by id, name, and label
    by_id: dict[str, ScorerResponse] = {str(s.id): s for s in all_resolved}
    by_name: dict[str, ScorerResponse] = {s.name: s for s in all_resolved}
    by_label: dict[str, ScorerResponse] = {}
    for s in all_resolved:
        label = s.label if not isinstance(s.label, type(None)) else None
        if label:
            by_label[label] = s

    scorers: list[ScorerConfig] = []
    unknown_metrics: list[str] = []

    # Match scorer_ids
    for sid in scorer_ids:
        matched = by_id.get(sid)
        if matched:
            scorers.append(ScorerConfig.from_dict(matched.to_dict()))
        else:
            unknown_metrics.append(sid)

    # Match label searches
    for search_val, version in label_searches:
        # Fall back to name (deprecated) match if label match not found
        matched = by_label.get(search_val) or by_name.get(search_val)
        if matched:
            raw_metric_dict = matched.to_dict()
            if version is not None:
                raw_version = scorers_client.get_scorer_version(scorer_id=raw_metric_dict["id"], version=version)
                raw_metric_dict["scorer_version"] = raw_version.to_dict()
            scorers.append(ScorerConfig.from_dict(raw_metric_dict))
        else:
            unknown_metrics.append(search_val)

    if unknown_metrics:
        raise ValueError(
            "One or more non-existent metrics are specified: " + ", ".join(f"'{metric}'" for metric in unknown_metrics)
        )

    # Register server-side metrics with Galileo.
    # Skip registration when run_id is None (trigger=True flow — API handles it).
    if run_id is not None and scorers:
        ScorerSettings().create(project_id=project_id, run_id=run_id, scorers=scorers)

    return scorers, local_metric_configs



import warnings
from collections.abc import Iterator
from enum import Enum
from typing import Any, Callable, Generic, Optional, TypeVar, Union

from pydantic import BaseModel, Field, ValidationError, field_validator
from pydantic_core.core_schema import ValidationInfo

from galileo_core.schemas.logging.span import Span
from galileo_core.schemas.logging.step import STEP_TYPES_WITH_CHILD_SPANS, StepType
from galileo_core.schemas.logging.trace import Trace
from galileo_core.schemas.shared.metric import MetricValueType
from galileo_core.schemas.shared.scorers.scorer_name import ScorerName


class GalileoMetrics(str, Enum):
    """Built-in Galileo metric scorers.

    Values are human-readable UI labels used for scorer lookup via the API.
    Member names follow the convention: base name = LLM version, _luna suffix = SLM version.
    """

    action_advancement = "Action Advancement"
    action_advancement_luna = "Action Advancement (SLM)"
    action_completion = "Action Completion"
    action_completion_luna = "Action Completion (SLM)"
    agent_efficiency = "Agent Efficiency"
    agent_flow = "Agent Flow"
    chunk_attribution_utilization = "Chunk Attribution Utilization"
    chunk_attribution_utilization_luna = "Chunk Attribution Utilization (SLM)"
    chunk_relevance = "Chunk Relevance"
    completeness = "Completeness"
    completeness_luna = "Completeness (SLM)"
    context_adherence = "Context Adherence"
    context_adherence_luna = "Context Adherence (SLM)"
    context_precision = "Context Precision"
    context_relevance = "Context Relevance"
    context_relevance_luna = "Context Relevance (SLM)"
    conversation_quality = "Conversation Quality"
    correctness = "Correctness"
    ground_truth_adherence = "Ground Truth Adherence"
    input_pii = "Input PII"
    input_pii_luna = "Input PII (SLM)"
    input_sexism = "Input Sexism"
    input_sexism_luna = "Input Sexism (SLM)"
    input_tone = "Input Tone"
    input_tone_luna = "Input Tone (SLM)"
    input_toxicity = "Input Toxicity"
    input_toxicity_luna = "Input Toxicity (SLM)"
    instruction_adherence = "Instruction Adherence"
    output_pii = "Output PII"
    output_pii_luna = "Output PII (SLM)"
    output_sexism = "Output Sexism"
    output_sexism_luna = "Output Sexism (SLM)"
    output_tone = "Output Tone"
    output_tone_luna = "Output Tone (SLM)"
    output_toxicity = "Output Toxicity"
    output_toxicity_luna = "Output Toxicity (SLM)"
    precision_at_k = "Precision@K"
    prompt_injection = "Prompt Injection"
    prompt_injection_luna = "Prompt Injection (SLM)"
    reasoning_coherence = "Reasoning Coherence"
    sql_adherence = "SQL Adherence"
    sql_correctness = "SQL Correctness"
    sql_efficiency = "SQL Efficiency"
    sql_injection = "SQL Injection"
    tool_error_rate = "Tool Error Rate"
    tool_error_rate_luna = "Tool Error Rate (SLM)"
    tool_selection_quality = "Tool Selection Quality"
    tool_selection_quality_luna = "Tool Selection Quality (SLM)"
    user_intent_change = "User Intent Change"


class _GalileoScorersProxyMeta(type):
    """Metaclass that makes GalileoScorers a deprecated proxy for ScorerName."""

    _DEPRECATION_MSG = (
        "GalileoScorers is deprecated and will be removed in a future release. Use galileo_core ScorerName instead."
    )

    def __getattribute__(cls, name: str) -> Any:
        if name.startswith("_") or name in ("__class__", "__mro__", "__dict__", "__bases__"):
            return type.__getattribute__(cls, name)
        warnings.warn(cls._DEPRECATION_MSG, DeprecationWarning, stacklevel=2)
        return getattr(ScorerName, name)

    def __getitem__(cls, name: str) -> ScorerName:
        warnings.warn(cls._DEPRECATION_MSG, DeprecationWarning, stacklevel=2)
        return ScorerName[name]

    def __instancecheck__(cls, instance: Any) -> bool:
        return isinstance(instance, ScorerName)

    def __subclasscheck__(cls, subclass: type) -> bool:
        return issubclass(subclass, ScorerName)

    def __iter__(cls) -> Iterator[ScorerName]:
        warnings.warn(cls._DEPRECATION_MSG, DeprecationWarning, stacklevel=2)
        return iter(ScorerName)

    def __contains__(cls, item: Any) -> bool:
        warnings.warn(cls._DEPRECATION_MSG, DeprecationWarning, stacklevel=2)
        return item in ScorerName

    def __dir__(cls) -> list[str]:
        return [m.name for m in ScorerName]

    def __repr__(cls) -> str:
        return "<deprecated GalileoScorers proxy - use ScorerName>"


class _GalileoScorersProxy(metaclass=_GalileoScorersProxyMeta):
    """Deprecated proxy that forwards to ScorerName."""

    def __new__(cls, value: Any) -> ScorerName:
        warnings.warn(_GalileoScorersProxyMeta._DEPRECATION_MSG, DeprecationWarning, stacklevel=2)
        return ScorerName(value)


GalileoScorers = _GalileoScorersProxy


MetricType = TypeVar("MetricType", bound=MetricValueType)


class LocalMetricConfig(BaseModel, Generic[MetricType]):
    name: str = Field(description="Name of the local metric")
    scorer_fn: Callable[[Union[Trace, Span]], MetricType] = Field(
        description="function to call to produce the metric value (takes a trace or span as input"
    )
    aggregator_fn: Optional[Callable[[list[MetricType]], Union[MetricType, dict[str, MetricType]]]] = Field(
        default=None,
        description="function to call to produce the aggregate metric values from individual metric values",
    )
    scorable_types: list[StepType] = Field(default=[StepType.llm])
    aggregatable_types: list[StepType] = Field(default=[StepType.trace])

    @field_validator("aggregatable_types", mode="before")
    def set_aggregatable_types(cls, value: list[StepType], info: ValidationInfo) -> list[StepType]:
        if "scorable_types" in info.data and any(
            scorable_type in value for scorable_type in info.data["scorable_types"]
        ):
            raise ValidationError("aggregatable_types cannot contain any types in scorable_types")
        for step_type in value:
            if step_type not in STEP_TYPES_WITH_CHILD_SPANS:
                raise ValidationError("aggregatable_types can only contain trace or workflow steps")
        return value


class Metric(BaseModel):
    name: str = Field(
        description="The name of the metric you want to run a specific version of (ie: 'Sentence Density')."
    )
    version: Optional[int] = Field(
        default=None,
        description="The version of the metric (ie: 1, 2, 3, etc.). If None is provided, the 'default' version will be used.",
    )



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.messages_list_item_role import MessagesListItemRole

if TYPE_CHECKING:
    from ..models.file_content_part import FileContentPart
    from ..models.text_content_part import TextContentPart


T = TypeVar("T", bound="MessagesListItem")


@_attrs_define
class MessagesListItem:
    """
    Attributes
    ----------
        content (Union[list[Union['FileContentPart', 'TextContentPart']], str]):
        role (Union[MessagesListItemRole, str]):
    """

    content: Union[list[Union["FileContentPart", "TextContentPart"]], str]
    role: Union[MessagesListItemRole, str]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.text_content_part import TextContentPart

        content: Union[list[dict[str, Any]], str]
        if isinstance(self.content, list):
            content = []
            for content_type_1_item_data in self.content:
                content_type_1_item: dict[str, Any]
                if isinstance(content_type_1_item_data, TextContentPart):
                    content_type_1_item = content_type_1_item_data.to_dict()
                else:
                    content_type_1_item = content_type_1_item_data.to_dict()

                content.append(content_type_1_item)

        else:
            content = self.content

        role: str
        role = self.role.value if isinstance(self.role, MessagesListItemRole) else self.role

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"content": content, "role": role})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.file_content_part import FileContentPart
        from ..models.text_content_part import TextContentPart

        d = dict(src_dict)

        def _parse_content(data: object) -> Union[list[Union["FileContentPart", "TextContentPart"]], str]:
            try:
                if not isinstance(data, list):
                    raise TypeError()
                content_type_1 = []
                _content_type_1 = data
                for content_type_1_item_data in _content_type_1:

                    def _parse_content_type_1_item(data: object) -> Union["FileContentPart", "TextContentPart"]:
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            return TextContentPart.from_dict(data)

                        except:  # noqa: E722
                            pass
                        if not isinstance(data, dict):
                            raise TypeError()
                        return FileContentPart.from_dict(data)

                    content_type_1_item = _parse_content_type_1_item(content_type_1_item_data)

                    content_type_1.append(content_type_1_item)

                return content_type_1
            except:  # noqa: E722
                pass
            return cast(Union[list[Union["FileContentPart", "TextContentPart"]], str], data)

        content = _parse_content(d.pop("content"))

        def _parse_role(data: object) -> Union[MessagesListItemRole, str]:
            try:
                if not isinstance(data, str):
                    raise TypeError()
                return MessagesListItemRole(data)

            except:  # noqa: E722
                pass
            return cast(Union[MessagesListItemRole, str], data)

        role = _parse_role(d.pop("role"))

        messages_list_item = cls(content=content, role=role)

        messages_list_item.additional_properties = d
        return messages_list_item

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Deprecated: use galileo.provider instead of galileo.__future__.provider."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.provider is deprecated. Use galileo.provider instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.provider import (  # noqa: E402
    AnthropicProvider,
    AzureProvider,
    BedrockProvider,
    GenericProvider,
    Model,
    OpenAIProvider,
    Provider,
    UnconfiguredProvider,
)

__all__ = [
    "AnthropicProvider",
    "AzureProvider",
    "BedrockProvider",
    "GenericProvider",
    "Model",
    "OpenAIProvider",
    "Provider",
    "UnconfiguredProvider",
]



from enum import Enum


class OrganizationAction(str, Enum):
    DELETE = "delete"
    DELETE_LOG_DATA = "delete_log_data"
    READ_SETTINGS = "read_settings"
    RENAME = "rename"
    UPDATE_SETTINGS = "update_settings"

    def __str__(self) -> str:
        return str(self.value)



"""Deprecated: use galileo.shared.utils instead of galileo.__future__.shared.utils."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.shared.utils is deprecated. Use galileo.shared.utils instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.shared.utils import classproperty  # noqa: E402

__all__ = ["classproperty"]



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.metric_aggregation import MetricAggregation

T = TypeVar("T", bound="MetricAggregationDetail")


@_attrs_define
class MetricAggregationDetail:
    """
    Attributes
    ----------
        id (str): Identifier for the metric in the response (e.g., 'w1', 'w2')
        metric_name (str): Name of the metric to aggregate
        aggregation (MetricAggregation):
    """

    id: str
    metric_name: str
    aggregation: MetricAggregation
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        metric_name = self.metric_name

        aggregation = self.aggregation.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"id": id, "metric_name": metric_name, "aggregation": aggregation})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        id = d.pop("id")

        metric_name = d.pop("metric_name")

        aggregation = MetricAggregation(d.pop("aggregation"))

        metric_aggregation_detail = cls(id=id, metric_name=metric_name, aggregation=aggregation)

        metric_aggregation_detail.additional_properties = d
        return metric_aggregation_detail

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="TextContentPart")


@_attrs_define
class TextContentPart:
    """A text segment within a message.

    Attributes
    ----------
        text (str):
        type_ (Union[Literal['text'], Unset]):  Default: 'text'.
    """

    text: str
    type_: Union[Literal["text"], Unset] = "text"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        text = self.text

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"text": text})
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        text = d.pop("text")

        type_ = cast(Union[Literal["text"], Unset], d.pop("type", UNSET))
        if type_ != "text" and not isinstance(type_, Unset):
            raise ValueError(f"type must match const 'text', got '{type_}'")

        text_content_part = cls(text=text, type_=type_)

        text_content_part.additional_properties = d
        return text_content_part

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from galileo.exceptions import (
    AuthenticationError,
    BadRequestError,
    ConflictError,
    ForbiddenError,
    NotFoundError,
    RateLimitError,
    ServerError,
)
from galileo.utils.headers_data import get_sdk_header
from galileo_core.constants.request_method import RequestMethod
from galileo_core.helpers.api_client import ApiClient

from ... import errors
from ...models.http_validation_error import HTTPValidationError
from ...models.user_collaborator import UserCollaborator
from ...models.user_collaborator_create import UserCollaboratorCreate
from ...types import Response


def _get_kwargs(project_id: str, *, body: list["UserCollaboratorCreate"]) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": RequestMethod.POST,
        "return_raw_response": True,
        "path": f"/projects/{project_id}/users",
    }

    _kwargs["json"] = []
    for body_item_data in body:
        body_item = body_item_data.to_dict()
        _kwargs["json"].append(body_item)

    headers["Content-Type"] = "application/json"

    headers["X-Galileo-SDK"] = get_sdk_header()

    _kwargs["content_headers"] = headers
    return _kwargs


def _parse_response(
    *, client: ApiClient, response: httpx.Response
) -> Union[HTTPValidationError, list["UserCollaborator"]]:
    if response.status_code == 200:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = UserCollaborator.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200

    if response.status_code == 422:
        return HTTPValidationError.from_dict(response.json())

    # Handle common HTTP errors with actionable messages
    if response.status_code == 400:
        raise BadRequestError(response.status_code, response.content)
    if response.status_code == 401:
        raise AuthenticationError(response.status_code, response.content)
    if response.status_code == 403:
        raise ForbiddenError(response.status_code, response.content)
    if response.status_code == 404:
        raise NotFoundError(response.status_code, response.content)
    if response.status_code == 409:
        raise ConflictError(response.status_code, response.content)
    if response.status_code == 429:
        raise RateLimitError(response.status_code, response.content)
    if response.status_code >= 500:
        raise ServerError(response.status_code, response.content)
    raise errors.UnexpectedStatus(response.status_code, response.content)


def _build_response(
    *, client: ApiClient, response: httpx.Response
) -> Response[Union[HTTPValidationError, list["UserCollaborator"]]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    project_id: str, *, client: ApiClient, body: list["UserCollaboratorCreate"]
) -> Response[Union[HTTPValidationError, list["UserCollaborator"]]]:
    """Create User Project Collaborators.

     Share a project with users.

    Args:
        project_id (str):
        body (list['UserCollaboratorCreate']):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Response[Union[HTTPValidationError, list['UserCollaborator']]]
    """
    kwargs = _get_kwargs(project_id=project_id, body=body)

    response = client.request(**kwargs)

    return _build_response(client=client, response=response)


def sync(
    project_id: str, *, client: ApiClient, body: list["UserCollaboratorCreate"]
) -> Optional[Union[HTTPValidationError, list["UserCollaborator"]]]:
    """Create User Project Collaborators.

     Share a project with users.

    Args:
        project_id (str):
        body (list['UserCollaboratorCreate']):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Union[HTTPValidationError, list['UserCollaborator']]
    """
    return sync_detailed(project_id=project_id, client=client, body=body).parsed


async def asyncio_detailed(
    project_id: str, *, client: ApiClient, body: list["UserCollaboratorCreate"]
) -> Response[Union[HTTPValidationError, list["UserCollaborator"]]]:
    """Create User Project Collaborators.

     Share a project with users.

    Args:
        project_id (str):
        body (list['UserCollaboratorCreate']):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Response[Union[HTTPValidationError, list['UserCollaborator']]]
    """
    kwargs = _get_kwargs(project_id=project_id, body=body)

    response = await client.arequest(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    project_id: str, *, client: ApiClient, body: list["UserCollaboratorCreate"]
) -> Optional[Union[HTTPValidationError, list["UserCollaborator"]]]:
    """Create User Project Collaborators.

     Share a project with users.

    Args:
        project_id (str):
        body (list['UserCollaboratorCreate']):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Union[HTTPValidationError, list['UserCollaborator']]
    """
    return (await asyncio_detailed(project_id=project_id, client=client, body=body)).parsed



"""Deprecated: use galileo.metric instead of galileo.__future__.metric."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.metric is deprecated. Use galileo.metric instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.metric import BuiltInMetrics, CodeMetric, GalileoMetric, LlmMetric, LocalMetric, Metric  # noqa: E402

__all__ = ["BuiltInMetrics", "CodeMetric", "GalileoMetric", "LlmMetric", "LocalMetric", "Metric"]



"""Deprecated: use galileo.shared.query_result instead of galileo.__future__.shared.query_result."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.shared.query_result is deprecated. Use galileo.shared.query_result instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.shared.query_result import QueryResult, _flatten_dict  # noqa: E402

__all__ = ["QueryResult", "_flatten_dict"]



"""Re-export from galileo.shared — will be deprecated once all __future__ modules are migrated."""

from galileo.shared.filter import boolean, date, number, text
from galileo.shared.sort import sort

__all__ = ["boolean", "date", "number", "sort", "text"]



"""Deprecated: use galileo.shared.sort instead of galileo.__future__.shared.sort."""

import warnings

warnings.warn(
    "Importing from galileo.__future__.shared.sort is deprecated. Use galileo.shared.sort instead.",
    DeprecationWarning,
    stacklevel=2,
)

from galileo.shared.sort import Sort, sort  # noqa: E402

__all__ = ["Sort", "sort"]
